from awsgnssroutils.database import OccList
from datetime import datetime 
from .timestandards import Time

awsro_time_convention = "utc"

#  Exception handling. 

class Error( Exception ): 
    pass

class awsroError( Error ): 
    def __init__( self, message, comment ): 
        self.message = message
        self.comment = comment


def get_occ_times( occs ): 
    """Generate a list of instance of timestandards.Time of the occultations in occs. 
    occs must be an instance of OccList."""

    if not isinstance(occs,OccList): 
        raise awsroError( "InvalidArgument", "Argument must be an instance of OccList" )

    datetimes = occs.values( "datetime" )
    ret = [ Time( **{ awsro_time_convention: datetime.strptime( t, "%Y-%m-%d-%H-%M" ) } ) for t in datetimes ]

    #  Done. 

    return ret




"""eumetsat.py


Author: Stephen Leroy (sleroy@aer.com)
Date: February 29, 2024

This module contains classes and methods useful for accessing the EUMETSAT Data Store. 
Note that you'll have to create an account on the EUMETSAT Data Store beforehand, set 
credentials using "eumdac set-credentials" at the Linux command line, and then set the 
default download root path using the setdefaults method below.
"""


import os, re, stat, json, subprocess 
from eumdac.token import AccessToken
from eumdac.datastore import DataStore
from datetime import datetime, timedelta
from .timestandards import Time
from .constants_and_utils import defaults_file


#  Definitions. 

metop_satellites = [ "Metop-A", "Metop-B", "Metop-C" ]
eumetsat_time_convention = "utc"
HOME = os.path.expanduser( "~" )
root_path_variable = "eumetsat_data_store_root"

#  Exception handling. 

class Error( Exception ): 
    pass 

class eumetsatError( Error ): 
    def __init__( self, message, comment ): 
        self.message = message
        self.comment = comment 


def setdefaults( root_path=None, eumetsattokens=None ): 
    """This method sets the default root path for EUMETSAT Data Store downloads.
    It also can set the consumer key and consumer secret as created on the 
    EUMETSAT Data Store website/account. If the keyword eumetsattokens is given, 
    it must be a 2-element tuple/list containing the user's consumer key and 
    consumer secret as provided for the user's account on the EUMETSAT Data 
    Store."""

    ret = { 'status': None, 'messages': [], 'comments': [] }

    if isinstance(root_path,str): 

        if os.path.exists( defaults_file ): 
            with open( defaults_file, 'r' ) as f: 
                defaults = json.load( f )
        else: 
            defaults = {}

        path = os.path.abspath( os.path.expanduser( root_path ) ) 
        defaults.update( { root_path_variable: path } )
        os.makedirs( path, exist_ok=True )

        #  Write to defaults file. 

        with open( defaults_file, 'w' ) as f: 
            json.dump( defaults, f, indent="  " )
        os.chmod( defaults_file, stat.S_IRUSR | stat.S_IWUSR )

    #  Define consumer key, consumer secret. 

    if eumetsattokens is not None: 
        if isinstance(eumetsattokens,tuple) or isinstance(eumetsattokens,list): 
            if len(eumetsattokens) == 2: 
                command = [ "eumdac", "set-credentials", eumetsattokens[0], eumetsattokens[1] ]
                resp = subprocess.run( command, capture_output=True )

            else:
                ret['status'] = "fail"
                ret['messages'].append( "InvalidArgument" )
                ret['comments'].append( "eumetsattokens must have length 2" )
                return ret

        else:
            ret['status'] = "fail"
            ret['messages'].append( "InvalidArgument" )
            ret['comments'].append( "eumetsattokens must be a tuple or a list" )
            return ret

    #  Done. 

    ret['status'] = "success"

    return


def checkdefaults(): 
    """Check that all of the defaults needed for execution have been set."""

    ret = { 'status': None, 'messages': [], 'comments': [] }

    #  Check for existence of ~/.eumdac, which is created by the eumdac --set-credentials command. 

    HOME = os.path.expanduser( "~" )
    credentials_file = os.path.join( HOME, ".eumdac", "credentials" )
    if not os.path.exists( credentials_file ): 
        ret['status'] = "fail"
        ret['messages'].append( "MissingCredentialsFile" )
        ret['comments'].append( 'Credentials file for the EUMETSAT Data Store is missing; be certain to run "rotcol setdefaults eumetsat ..."' )
        return ret

    #  Check for existence of package defaults file. 

    if not os.path.exists( defaults_file ): 
        ret['status'] = "fail"
        ret['messages'].append( "MissingDefaultsFile" )
        ret['comments'].append( 'Missing defaults file; be certain to run "rotcol setdefaults ..."' )
        return ret

    #  Read defaults file. 

    with open( defaults_file, 'r' ) as f: 
        defaults = json.load( f )

    #  Check that the data root path has been set. 

    if root_path_variable not in defaults.keys(): 
        ret['status'] = "fail"
        ret['messages'].append( "MissingEUMETSATDataStoreRoot" )
        ret['comments'].append( 'Missing data root for EUMETSAT Data Store; be certain to run "rotcol setdefaults eumetsat --dataroot ..."' )
        return ret

    #  Done. 

    ret['status'] = "success"
    return ret


class EUMETSATDataStore(): 
    """Class to handle interaction with the EUMETSAT Data Store."""

    def __init__( self ): 
        """Create an instance of EUMETSATDataStore. data_root should be an 
        absolute path to the storage location for downloaded EUMETSAT Data
        Store data."""

        #  Get credentials. 

        credentials_file = os.path.join( HOME, ".eumdac", "credentials" )
        if not os.path.exists( credentials_file ): 
            raise eumetsatError( "CredentialsFileNotFound", 
                    "Be sure to run eumdac set-credentials on the command line" )

        f = open( credentials_file, 'r' ) 
        self.credentials = tuple( f.readline().split(",") )
        f.close()

        self.token = None
        self.datastore = None

        #  Get root data path. 

        with open( defaults_file, 'r' ) as f: 
            defaults = json.load( f )
            self.data_root = defaults[root_path_variable]

        self.inventory = {}

        #  Initialize token, inventory. 

        self.get_token()
        self.inventory_metop_amsua()

    def get_token( self ): 
        self.token = AccessToken( self.credentials )
        self.datastore = DataStore( self.token )

    def inventory_metop_amsua( self ): 
        """Create an inventory of the files available on the local file system, 
        keeping information on the absolute path and the time range of data."""

        #  Get list of satellites. 

        data_root = os.path.join( self.data_root, "amsua" )
        satellites = [ p for p in os.listdir( data_root ) if re.search( "^Metop-", p ) \
                and os.path.isdir( os.path.join( data_root, p ) ) ]

        #  Initialize inventory. 

        self.inventory = { sat: [] for sat in satellites }

        #  Loop over satellites. 

        for sat in satellites: 

            for root, subdirs, files in os.walk( os.path.join( data_root, sat ) ): 
                subdirs.sort()
                files.sort()

                ss = "^" + sat + "_AMSAL1_(\d{8}T\d{6})Z_(\d{8}T\d{6})Z\.nc$"
                st = "%Y%m%dT%H%M%S" 

                for file in files: 
                    m = re.search( ss, file )
                    if m is None: 
                        continue
                    t1 = Time( **{ eumetsat_time_convention: datetime.strptime( m.group(1), st ) } ) 
                    t2 = Time( **{ eumetsat_time_convention: datetime.strptime( m.group(2), st ) } ) 

                    rec = { 'satellite': sat, 'path': os.path.join( root, file ), 'timerange': ( t1, t2 ) }
                    self.inventory[sat].append( rec )

    def get_metop_amsua_paths( self, satellite, timerange ): 
        """Return a listing of the paths to Metop AMSU-A data files given a 
        satellite name and a time range. The timerange is a two-element 
        tuple/list with instances of timestandards.Time or datetime.datetime. 
        If it is the latter, then the datetime elements are understood to be 
        UTC."""

        #  Check input. Interpret datetime.datetime as timestandards.Time instances 
        #  if necessary. 

        if len( timerange ) != 2: 
            raise eumetsatError( "InvalidArgument", "timerange must be a tuple/list of two elements" )

        if isinstance( timerange[0], datetime ) and isinstance( timerange[1], datetime ): 
            _timerange = [ Time( utc=timerange[i] ) for i in range(2) ]
        elif isinstance( timerange[0], Time ) and isinstance( timerange[1], Time ): 
            _timerange = timerange
        else: 
            raise eumetsatError( "InvalidArgument", "The elements of timerange must both be " + \
                    "datetime.datetime or timestandards.Time" )

        ret = sorted( [ rec['path'] for rec in self.inventory[satellite] if \
                rec['timerange'][0] <= _timerange[1] and rec['timerange'][1] >= _timerange[0] ] )

        return ret

    def populate_metop_amsua( self, satellite, timerange ): 
        """Download a Metop AMSU-A data that falls within a timerange. 

        * satellite must be one of Metop-A, Metop-B, Metop-C. 
        * timerange is a 2-element tuple/list of instances of timestandards.Time 
          or instances of datetime.datetime defining the range of times over which 
          to retrieve data. If they are instances of datetime.datetime, then the 
          convention is that they are both UTC."""

        #  Check input. Interpret datetime.datetime as timestandards.Time instances 
        #  if necessary. 

        if len( timerange ) != 2: 
            raise eumetsatError( "InvalidArgument", "timerange must be a tuple/list of two elements" )

        if isinstance( timerange[0], datetime ) and isinstance( timerange[1], datetime ): 
            _timerange = [ Time( utc=timerange[i] ) for i in range(2) ]
        elif isinstance( timerange[0], Time ) and isinstance( timerange[1], Time ): 
            _timerange = timerange
        else: 
            raise eumetsatError( "InvalidArgument", "The elements of timerange must both be " + \
                    "datetime.datetime or timestandards.Time" )

        #  Load the data collection interface. 

        amsua_collection = self.datastore.get_collection( "EO:EUM:DAT:METOP:AMSUL1" )

        #  Establish time interval and get inventory of data at the EUMETSAT Data 
        #  Store

        start = _timerange[0].calendar(eumetsat_time_convention).isoformat()
        end = _timerange[1].calendar(eumetsat_time_convention).isoformat()

        command = [ "eumdac", "search", "-c", "EO:EUM:DAT:METOP:AMSUL1", "--satellite",  
                satellite, "-s", start, "-e", end ]

        if datetime.now() >= self.token.expiration: 
            self.get_token()

        ret = subprocess.run( command, capture_output=True )
        inventory = sorted( ret.stdout.decode().split("\n") )

        #  Scan inventory of files; check to see if all files already exist in data 
        #  repository directory. 

        download = False
        matched_files = [] 

        for item in inventory: 

            m = re.search( "(\d{14})Z_(\d{14})Z", item )
            if not m: continue

            #  Start time and end time of inventory item. 

            dt1 = datetime.strptime( m.group(1), "%Y%m%d%H%M%S" )
            dt2 = datetime.strptime( m.group(2), "%Y%m%d%H%M%S" )

            #  Does the directory corresponding to this inventory item exist on the 
            #  local file system? If not, then download all data. 

            dirpath = os.path.join( self.data_root, "amsua", satellite, f'{dt1.year:4d}', f'{dt1.month:02d}', f'{dt1.day:02d}' )
            if not os.path.isdir( dirpath ): 
                download = True
                break

            #  Get listing of all existing files on for this day. 

            local_files = os.listdir( dirpath )
            ss = dt1.strftime( "%Y%m%dT%H%M%SZ" ) + "_" + dt2.strftime( "%Y%m%dT%H%M%SZ" ) 
            match = [ file for file in local_files if re.search( ss, file ) ]

            #  Does a matched file exist? If not, download all files. 

            if len(match) == 0: 
                download = True
                break
            elif len(match) == 1: 
                matched_files.append( os.path.join( dirpath, match[0] ) )

        #  Download data if necessary. 

        if download: 

            os.makedirs( "tmp", exist_ok=True )

            #  Flush tmp holding directory. 

            existing_files = os.listdir( "tmp" )
            for file in existing_files: 
                os.unlink( os.path.join( "tmp", file ) )

            #  Initialize and enter loop of download chunks. Each download chunk is limited 
            #  to 5 days. 

            current_start_time = _timerange[0] + 0
            outpaths = []

            while current_start_time < _timerange[1]: 

                #  Establish time interval for download. 

                current_end_time = min( [ current_start_time + 86400 * 5, _timerange[1] ] )

                #  Define download command. 

                current_start = current_start_time.calendar(eumetsat_time_convention).isoformat()
                current_end = current_end_time.calendar(eumetsat_time_convention).isoformat()

                command = [ "eumdac", "download", "--collection", "EO:EUM:DAT:METOP:AMSUL1",
                        "--satellite", satellite, "-s", current_start, "-e", current_end, "-o", "tmp", "-y",
                        "--chain", f"product: AMSAL1, format: netcdf4" ]

                #  Renew access token if expired. 

                if datetime.now() >= self.token.expiration: 
                    self.get_token()

                ret = subprocess.run( command, capture_output=True )

                #  Scan the downloaded files, establish repository path according to 
                #  time of the first sounding in each file, and transfer them into the 
                #  repository. 

                files = os.listdir( "tmp" )

                for file in files:
                    inpath = os.path.join( "tmp", file )
                    m = re.search( "^AMSAL1_(\d{8}T\d{6})Z_(\d{8}T\d{6})Z_epct_.*\.nc$", file )
                    t = datetime.strptime( m.group(1), "%Y%m%dT%H%M%S" )
                    outpath = os.path.join( self.data_root, "amsua", satellite, 
                            f'{t.year:4d}', f'{t.month:02d}', f'{t.day:02d}',
                            f'{satellite}_AMSAL1_{m.group(1)}Z_{m.group(2)}Z.nc' )
                    if not os.path.exists( outpath ):
                        os.makedirs( os.path.dirname(outpath), exist_ok=True )
                        os.link( inpath, outpath )
                    os.unlink( inpath )
                    outpaths.append( outpath )

                #  Next 5-day chunk...

                current_start_time += 86400 * 5

            #  Exit loop over 5-day chunks.

            #  Refresh inventory. 

            self.inventory_metop_amsua()

        #  Done. 

        ret = self.get_metop_amsua_paths( satellite, timerange )

        return ret




"""database.py

Authors: Amy McVey (amcvey@aer.com) and Stephen Leroy (sleroy@aer.com)
Date: 10 January 2024

================================================================================

This module contains utilities to query the AWS Registry of Open Data metadata_root
of GNSS radio occultation data. It does so using database files posted in the
AWS metadata_root.

Functionality
=============
This module defines two classes: RODatabaseClient and OccList. The first
creates a portal to a database of RO metadata, and the second is an instance
of a list of radio occultations (ROs). Each are described below.

RODatabaseClient:
    Create an instance of a portal to a metadata on all RO data in the AWS
    Registry of Open Data. It provides an option to create a metadata_root of
    the RO metadata on the local file system as keyword "metadata_root".

OccList:
    An instance of the class OccList is contains the metadata on a list of RO
    soundings along with pointers to the RO data files in the AWS Registry of
    Open Data S3 bucket. AWS functionality is completely embedded in the
    methods of the OccList class. Those methods include the ability to
    subset/filter the list according to geolocation and time,
    GNSS transmitter/constellation, GNSS receiver, whether it is a rising or a
    setting occultation, etc. It also includes the ability to combine
    instances of OccList, save the OccList to a JSON format file for future
    restoration by RODatabaseClient.restore, and even download RO data files.

Very useful utilities...

setdefaults: 
    A function that sets defaults for use of RODatabaseClient and OccList. 
    It allows the user to specify defaults for storage of RO metadata 
    database files ("metadata_root") and for downloads of RO data files 
    ("data_root"). In doing so, the user won't have to specify the metadata_root 
    every time an instance of RODatabaseClient is created nor specify a 
    data download path every time the OccList.download method is called. 

populate:
    Pre-populate the metadata database in the default metadata storage 
    directory. This will greatly accelerate all queries of the database, 
    including first queries. Be sure to have run setdefaults in advance. 

See README documentation for further instruction on usage. 
"""

#  Define parameters of the database.

AWSregion = "us-east-1"
databaseS3bucket = "gnss-ro-data"
float_fill_value = -999.99
defaults_filename = ".awsgnssroutilsrc"

#  Imports.

import os
import datetime
import numpy as np
import boto3
import json
import re
import time
from tqdm import tqdm
import subprocess
from botocore import UNSIGNED


#  Linux epoch. 

linux_epoch = datetime.datetime( 1970, 1, 1, tzinfo=datetime.timezone.utc )

#  Exception handling.

class Error( Exception ):
    pass

class AWSgnssroutilsError( Error ):
    def __init__( self, message, comment ):
        self.message = message
        self.comment = comment


################################################################################
#  Useful utility functions and classes.
################################################################################

def unsigned_S3Client():
    """This is a custom function that contains code to generate an authenticated
    instance of a boto3 S3 client. In this particular case, authentication is
    UNSIGNED."""

    session = boto3.Session( region_name=AWSregion )
    s3client = session.client( "s3", config = boto3.session.Config( signature_version=UNSIGNED ) )

    return s3client

class S3Wrapper():
    """This class is a wrapper for a boto3 s3 cleint that checks for broken
    instances and restores them when necessary."""

    def __init__( self, S3Client_create_function, bucket ):
        """Create a wrapper for a boto3 s3 client. The sole argument points to a
        function that returns an instance of boto3.Session.client('s3') that is 
        fully authenticated."""

        self._s3clientcreate = S3Client_create_function
        message = "Argument to S3Client must be a reference to a function " + \
                "that returns an instance of a boto3 s3 client." 

        try:
            self._s3client = self._s3clientcreate()
        except:
            raise AWSgnssroutilsError( "IncorrectArgument", message )

        self.bucket = bucket

    def info( self, prefix ):
        try:
            ret = self._s3client.list_objects_v2( Bucket=self.bucket, Prefix=prefix )
        except:
            self._s3client = self._s3clientcreate()
            ret = self._s3client.list_objects_v2( Bucket=self.bucket, Prefix=prefix )
        return ret[0]

    def download( self, prefix, y ):
        try:
            ret = self._s3client.download_file( self.bucket, prefix, y )
        except:
            self._s3client = self._s3clientcreate()
            ret = self._s3client.download_file( self.bucket, prefix, y )
        return ret

    def ls( self, prefix ):

        try: 
            paginator = self._s3client.get_paginator( "list_objects_v2" )
        except:
            self._s3client = self._s3clientcreate()
            paginator = self._s3client.get_paginator( "list_objects_v2" )

        pages = paginator.paginate( Bucket=self.bucket, Prefix=prefix, Delimiter="/" )
        ret = { 'prefixes':[], 'keys':[] }

        for page in pages: 
            if "CommonPrefixes" in page.keys(): 
                ret['prefixes'] += [ m['Prefix'] for m in page['CommonPrefixes'] ]
            if "Contents" in page.keys(): 
                ret['keys'] += [ m['Key'] for m in page['Contents'] ]

        return ret

################################################################################
#  IMPORTANT!!!!
#  Define the auto-refresh S3 client that will be used. 
################################################################################

use_S3Client = unsigned_S3Client 

################################################################################


#  Useful parameters. Scan the AWS RO data metadata_root for valid versions 
#  (valid_versions), valid processing centers (valid_processing_centers), 
#  valid file types (valid_file_types), and valid missions (valid_missions). 
#
#  valid_versions -> a list of version identifiers
#
#  valid_processing_centers[version] -> a list of processing centers for a 
#           specified version
#
#  valid_file_types[version] -> a list of existing file types for a 
#           specified version
#
#  valid_missions[version] -> a list of missions for a specified version
#
#  validity_table -> a list of dictionaries defining the versions, processing 
#           centers, file types, and missions available. 
#

s3client = use_S3Client()
kwargs = { 'Bucket': databaseS3bucket, 'Delimiter': "/" }

#  Create valid versions. 

prefixes = s3client.list_objects_v2( Prefix=f'contributed/', **kwargs )['CommonPrefixes'] 
valid_versions = [ entry['Prefix'].split("/")[-2] for entry in prefixes ]

#  Initialize bookkeeping for loops over versions, centers, and missions. 

valid_processing_centers = {}
valid_file_types = {}
valid_missions = {}
valid_table = []

for version in valid_versions: 

    prefixes = s3client.list_objects_v2( Prefix=f'contributed/{version}/', **kwargs )['CommonPrefixes'] 
    valid_processing_centers.update( { version: list( { entry['Prefix'].split("/")[-2] for entry in prefixes } ) } )
    valid_missions.update( { version: [] } )

    all_filetypes = []

    for center in valid_processing_centers[version]: 

        prefixes = s3client.list_objects_v2( Prefix=f'contributed/{version}/{center}/', **kwargs )['CommonPrefixes'] 
        missions = [ entry['Prefix'].split("/")[-2] for entry in prefixes ]
        valid_missions[version] += missions

        for mission in missions: 

            prefixes = s3client.list_objects_v2( Prefix=f'contributed/{version}/{center}/{mission}/', **kwargs )['CommonPrefixes'] 
            filetypes = [ entry['Prefix'].split("/")[-2] for entry in prefixes ]

            all_filetypes += [ entry['Prefix'].split("/")[-2] for entry in prefixes ]
            valid_table += [ { 'version': version, 'center': center, 'mission': mission, 'filetype': ft } for ft in filetypes ] 

    valid_file_types.update( { version: list( set( all_filetypes ) ) } )

valid_missions = { version: sorted( list( set( missions ) ) ) for version, missions in valid_missions.items() }


################################################################################
#  Resources utilities: Create a defaults file, populate metadata database. 
################################################################################

def setdefaults( metadata_root:str=None, data_root:str=None, version:str=None ) -> dict: 
    """Create a local metadata_root for a history database queries. 

    Create a local metadata_root of prior database queries and record the 
    directory path in a defaults file that can be found in the user's home 
    directory. Also, one can specify a default path for RO data downloads. 

    metadata_root       An absolute path to a directory containing 
                        RO metadata information from all previous searches. 

    data_root           The default path for downloading RO data. If 
                        it is not provided here, then it must be provided 
                        when the method OccList.download is called. "data_root" 
                        must be an absolute path.

    version             The default RO format version. The available versions 
                        can be found in s3://gnss-ro-data/dynamo. 
    """

    # defaults = { 'metadata_root': "", 'data_root': "", 'version': "" }
    new_defaults = {}

    #  Check for existence of defaults file. Get existing defaults. 

    HOME = os.path.expanduser( "~" )
    defaults_file_path = os.path.join( HOME, defaults_filename )

    if os.path.exists( defaults_file_path ): 
        with open( defaults_file_path, 'r' ) as fp: 
            defaults = json.load( fp )
    else: 
        defaults = {}

    #  Be sure the paths are absolute paths. Also, create paths if they 
    #  don't already exist. 

    if metadata_root is not None: 
        if metadata_root != os.path.abspath( metadata_root ): 
            raise AWSgnssroutilsError( "BadPathName", 
                    f'Path to metadata_root ({metadata_root}) must be an absolute path' )
        else: 
            defaults.update( { 'metadata_root': metadata_root } )
            os.makedirs( metadata_root, exist_ok=True )

    if data_root is not None: 
        if data_root != os.path.abspath( data_root ): 
            raise AWSgnssroutilsError( "BadPathName", 
                    f'Path to data_root ({data_root}) must be an absolute path' )
        else: 
            defaults.update( { 'data_root': data_root } )
            os.makedirs( data_root, exist_ok=True )


    #  Check for a valid version. 

    if version is not None: 

        if version not in valid_versions: 
            raise AWSgnssroutilsError( "InvalidVersion", f'Version "{version}" is invalid; ' + \
                    'valid versions are ' + ", ".join( valid_versions ) )

        defaults.update( { 'version': version } )

        if metadata_root is not None: 
            os.makedirs( os.path.join( metadata_root, version ), exist_ok=True )

    #  Record new set of defaults. 

    with open( defaults_file_path, 'w' ) as fp: 
        json.dump( defaults, fp )

    #  Done. 

    return defaults


def populate() -> subprocess.CompletedProcess : 
    """Populate the metadata database in the path established by 
    setdefaults. 

    This function will synchronize the default metadata_root path 
    "{respository}/{version}" with the contents in the AWS S3 path 
    s3://gnss-ro-data/dynamo/{version}/export_subsets. 
    """

    defaults = get_defaults()
    metadata_root, version = defaults['metadata_root'], defaults['version']

    if version is None: 
        raise AWSgnssroutilsError( "InvalidVersion", 'A default version must ' + \
                'be specified using the awsgnssroutils.database.setdefaults function.' )

    #  Check for validity of version. 

    if version not in valid_versions: 
        raise AWSgnssroutilsError( "InvalidVersion", f'Version "{version}" is invalid; ' + \
                'valid versions are ' + ", ".join( valid_versions ) )

    #  Create command to sync the metadata database. 

    command = [ 'aws', 's3', 'sync', 
               f's3://{databaseS3bucket}/dynamo/{version}/export_subsets/', 
               os.path.join(metadata_root,version), '--no-sign-request', '--delete' ]

    #  Synchronize using subprocess.

    ret = subprocess.run( command, capture_output=True )

    return


################################################################################
#  Internal utilities. 
################################################################################

def get_defaults() -> dict: 
    """Read the module defaults from a file in the user's home 
    directory and return contents in a dictionary."""

    #  Define defaults file path. 

    HOME = os.path.expanduser( "~" )
    defaults_file_path = os.path.join( HOME, defaults_filename )

    if not os.path.exists( defaults_file_path ): 
        raise AWSgnssroutilsError( "NoResourcesFile", 
                    f'The defaults file "{defaults_file_path}" could not be found. ' + \
                            'Be sure to create the defaults using the function ' + \
                            'awsgnssroutils.database.setdefaults.' )

    #  Read the defaults. 

    with open( defaults_file_path, 'r' ) as fp: 
        defaults = json.load( fp )

    #  Replace emptry strings with Nones. 

    for key, value in defaults.items(): 
        if value == "": defaults.update( { key: None } )

    #  Backward compatibility. 

    keys = defaults.keys()

    if "repository" in keys and "metadata_root" not in keys: 
        defaults.update( { 'metadata_root': defaults['repository'] } )

    if "rodata" in keys and "data_root" not in keys: 
        defaults.update( { 'data_root': defaults['rodata'] } )

    #  Done. 

    return defaults


################################################################################
#  Define the OccList class, which defines a list of occultations together with
#  the metadata on each occultation in the list.
################################################################################

class OccList():
    """An instance containing a list of entries/radio occultation soundings
    in the AWS Registry of Open Data for GNSS RO data. It provides methods
    to filter/subset the list, get metadata on the occultations in the list,
    save the list for future use, download relevant database metadata for
    future inquiries directly from the AWS Open Data S3 bucket, and download
    all RO data associated with the entries in the list to the local file
    system."""

    def __init__( self, data:list, s3wrapper:S3Wrapper, version:str ):
        """Create an instance of OccList. 

        Arguments
        =========
        data        A list of items/RO soundings from the RO database. 

        s3wrapper   An instance of S3Wrapper that provides access to the AWS 
                    metadata_root of RO data.

        version     The AWS metadata_root version.
        """

        if isinstance( data, list ):
            self._data = data
        else:
            raise AWSgnssroutilsError( "BadInput", "Input argument data must be a list." )

        if isinstance( s3wrapper, S3Wrapper ): 
            self._s3 = s3wrapper
        else: 
            raise AWSgnssroutilsError( "BadInput", "Second input argument must be " + \
                    "an instance of S3tWrapper." )

        self._version = version

        self.size = len( self._data )


    def filter( self, missions:{str,tuple,list}=None, 
               receivers:{str,tuple,list}=None, 
               transmitters:{str,tuple,list}=None,
               GNSSconstellations:{str,tuple,list}=None, 
               longituderange:{str,tuple,list}=None, 
               latituderange:{str,tuple,list}=None,
               datetimerange:{tuple,list}=None, 
               localtimerange:{tuple,list}=None, 
               geometry:str=None,
               availablefiletypes:{str,tuple,list}=None ):
        """Filter a list of occultations according to various criteria, such as
        mission, receiver, transmitter, etc.  

        Filtering can be done through the following keywords:

        missions            A string or list-like object containing the names of
                            the missions over which to apply the filter.

        receivers           A string or list-like object containing the names of
                            the GNSS RO receivers over which to apply the filter.

        transmitters        A string or list-like object containing the names of
                            the GNSS transmitters over which to apply the filter.

        GNSSconstellations  A string or list-like object containing the names of
                            the GNSS constellations ("G" for GPS, "R" for
                            GLONASS, etc.) over which to apply the filter.

        longituderange      A two-element list-like object containing the
                            longitude bounds of the soundings to retain upon
                            applying the filter. Longitudes are bounded by -180
                            to 180, and the region can wrap around the date line.

        latituderange       A two-element list-list object containing the
                            latitude bounds of the soundings to retain upon
                            applying the filter. Latitudes are bounded by -90 to
                            90.

        datetimerange       A two-element list-like object containing the
                            date-time bounds of the soundings to retain upon
                            applying the filter. Each of the two elements must
                            be a string object with an ISO-format date-time.

        localtimerange      A two-element list-like object containing the
                            local time bounds of the soundings to retain upon
                            applying the filter. Each of the two elements must
                            by a local/solar time in hours, ranging from 0 to
                            24. The range can wrap around midnight (0 hrs).

        geometry            A string defining the occultation geometry over
                            which to apply the filter. Must be "rising" to
                            restrict the list to rising occultations only, or
                            "setting" to restrict the list to setting
                            occultations only.

        availablefiletypes  A list, set, or tuple of strings designating the
                            AWS RO data file types that must be present for
                            the retained soundings. Each file type must be of
                            the format "{center}_{filetype}" where the "center"
                            is one of the valid contributing RO processing
                            centers ("ucar", "jpl", "romsaf", etc.) and the
                            "filetype" is one of the valid AWS RO data file
                            types ("calibratedPhase", "refractivityRetrieval",
                            "atmosphericRetrieval".
        """

        #  Filter by GNSSconstellations or by transmitters, but not by both.

        if GNSSconstellations is not None and transmitters is not None:
            raise AWSgnssroutilsError( "GNSSconstellationsTransmittersClash",
                    "Filtering by both GNSS constellation " + \
                    "and transmitter is not permitted." )

        #  Filter by mission or receivers, but not by both..

        if missions is not None and receivers is not None:
            raise AWSgnssroutilsError( "MissionsReceiversClash", "Filtering by both RO missions " + \
                    "and receiver is not permitted." )

        #  Check GNSSconstellations.

        if GNSSconstellations is not None:
            if isinstance( GNSSconstellations, str ):
                f_GNSSconstellations = [ GNSSconstellations ]
            elif isinstance( GNSSconstellations, tuple ) or \
                isinstance( GNSSconstellations, list ) or \
                isinstance( GNSSconstellations, np.ndarray ):
                    f_GNSSconstellations = list( GNSSconstellations )
            else:
                raise AWSgnssroutilsError( "FaultyGNSSconstellations", "GNSS constellations must be a " + \
                        "string or a list-like object" )
        else:
            f_GNSSconstellations = None

        #  Check transmitters.

        if transmitters is not None:
            if isinstance( transmitters, str ):
                f_transmitters = [ transmitters ]
            elif isinstance( transmitters, tuple ) or \
                isinstance( transmitters, list ) or \
                isinstance( transmitters, np.ndarray ):
                    f_transmitters = list( transmitters )
            else:
                raise AWSgnssroutilsError( "FaultyTransmitters", "transmitters must be a " + \
                        "string or a list-like object" )
        else:
            f_transmitters = None

        #  Check missions.

        if missions is not None:
            if isinstance( missions, str ):
                f_missions = [ missions ]
            elif isinstance( missions, tuple ) or \
                isinstance( missions, list ) or \
                isinstance( missions, np.ndarray ):
                    f_missions = list( missions )
            else:
                raise AWSgnssroutilsError( "FaultyMissions", "missions must be a " + \
                        "string or a list-like object" )
        else:
            f_missions = None

        #  Check receivers.

        if receivers is not None:
            if isinstance( receivers, str ):
                f_receivers = [ receivers ]
            elif isinstance( receivers, tuple ) or \
                isinstance( receivers, list ) or \
                isinstance( receivers, np.ndarray ):
                    f_receivers = list( receivers )
            else:
                raise AWSgnssroutilsError( "FaultyReceivers", "receivers must be a " + \
                        "string or a list-like object" )
        else:
            f_receivers = None

        #  Check longituderange.

        if longituderange is not None:
            if isinstance( longituderange, tuple ) or \
                isinstance( longituderange, list ) or \
                isinstance( longituderange, np.ndarray ):
                    if len( longituderange ) == 2:
                        f_longituderange = np.array( longituderange )
                        if np.logical_and( f_longituderange >= -180, f_longituderange <= 180 ).sum() != 2:
                            raise AWSgnssroutilsError( "FaultyLongitudeRange", "longitudes in " + \
                                    "longituderange must both fall between -180 and 180" )
                    else:
                        raise AWSgnssroutilsError( "FaultyLongitudeRange", "longituderange must have " + \
                                "two elements" )
            else:
                raise AWSgnssroutilsError( "FaultyLongitudeRange", "longituderange must be a tuple/list/ndarray" )
        else:
            f_longituderange = None

        #  Check latituderange.

        if latituderange is not None:
            if isinstance( latituderange, tuple ) or \
                isinstance( latituderange, list ) or \
                isinstance( latituderange, np.ndarray ):
                    if len( latituderange ) == 2:
                        f_latituderange = np.array( latituderange )
                        if np.logical_and( f_latituderange >= -90, f_latituderange <= 90 ).sum() != 2 or \
                            ( f_latituderange[1] <= f_latituderange[0] ):
                            raise AWSgnssroutilsError( "FaultyLatitudeRange", "latitudes in " + \
                                    "latituderange must both fall between -90 and 90 and " + \
                                    "latituderange[1] > latituderange[0]" )
                    else:
                        raise AWSgnssroutilsError( "FaultyLatitudeRange", "latituderange must have " + \
                                "two elements" )
            else:
                raise AWSgnssroutilsError( "FaultyLatitudeRange", "latituderange must be a tuple/list/ndarray" )
        else:
            f_latituderange = None

        #  Check datetimerange.

        if datetimerange is not None:
            if isinstance( datetimerange, tuple ) or \
                isinstance( datetimerange, list ) or \
                isinstance( datetimerange, np.ndarray ):
                    if len( datetimerange ) == 2:
                        try:
                            f_datetimerange = [ datetime.datetime.fromisoformat( datetimerange[0] ),
                                       datetime.datetime.fromisoformat( datetimerange[1] ) ]
                        except:
                            raise AWSgnssroutilsError( "FaultyDatetimerange", "The elements of datetimerange " + \
                                    "must be ISO format times" )
                        if f_datetimerange[0] > f_datetimerange[1]:
                            raise AWSgnssroutilsError( "FaultyDatetimerange", "datetimerange[0] " + \
                                    "must be less than datetimerange[1]" )
                    else:
                        raise AWSgnssroutilsError( "FaultyDatetimerange", "datetimerange must have " + \
                                "two elements" )
            else:
                raise AWSgnssroutilsError( "FaultyDatetimerange", "datetimerange must be a tuple/list" )
        else:
            f_datetimerange = None

        #  Check localtimerange.

        if localtimerange is not None:
            if isinstance( localtimerange, tuple ) or \
                isinstance( localtimerange, list ) or \
                isinstance( localtimerange, np.ndarray ):
                    if len( localtimerange ) == 2:
                        f_localtimerange = np.array( localtimerange )
                        if np.logical_and( f_localtimerange >= 0.0, f_localtimerange < 24.0 ).sum() != 2:
                            raise AWSgnssroutilsError( "FaultySolartimeRange", "localtimes in " + \
                                    "localtimerange must both fall between -180 and 180" )
                    else:
                        raise AWSgnssroutilsError( "FaultySolartimeRange", "localtimerange must have " + \
                                "two elements" )
            else:
                raise AWSgnssroutilsError( "FaultySolartimeRange", "localtimerange must be a tuple/list/ndarray" )
        else:
            f_localtimerange = None

        #  Check geometry.

        if geometry is not None:
            if geometry not in [ "setting", "rising" ]:
                raise AWSgnssroutilsError( "FaultyGeometry", "geometry can only be one of 'setting' or 'rising'" )
        f_geometry = geometry


        #  Check availablefiletypes.

        if availablefiletypes is not None:

            if type( availablefiletypes ) not in [ str, list, set, tuple ]:
                raise AWSgnssroutilsError( "FaultyAvailableFiletypes", 'availablefiletypes must be of class ' + \
                        '"str", "list", "set", or "tuple"' )

            if isinstance( availablefiletypes, str ):
                f_availablefiletypes = { availablefiletypes }
            else:
                f_availablefiletypes = set( availablefiletypes )

            for availablefiletype in f_availablefiletypes:
                m = re.search( r"^(\w+)_(\w+)$", availablefiletype )
                if m:
                    center, filetype = m.group(1), m.group(2)
                    if center not in valid_processing_centers[self._version]:
                        raise AWSgnssroutilsError( "InvalidProcessingCenter",
                                f'Processing center "{center}" in availablefiletype {availablefiletype} ' + \
                                        'is not valid.' )
                    if filetype not in valid_file_types[self._version]:
                        raise AWSgnssroutilsError( "InvalidFileType",
                                f'File type "{filetype}" in availablefiletype {availablefiletype} ' + \
                                        'is not valid.' )
                else:
                    raise AWSgnssroutilsError( "InvalidAvailableFiletype",
                                f'availablefiletype {availablefiletype} is not a valid format.' )
        else:
            f_availablefiletypes = None

        # Loop through list of items, each a dictionary.

        keep_list = []

        for item in self._data:

            keep = True

            if f_missions is not None:
                keep &= ( item['mission'] in f_missions )

            if f_receivers is not None:
                keep &= ( item['receiver'] in f_receivers )

            if f_GNSSconstellations is not None:
                keep &= ( item['transmitter'][0] in f_GNSSconstellations )

            if f_transmitters is not None:
                keep &= ( item['transmitter'] in f_transmitters )

            if f_datetimerange is not None:
                ms = re.match( r"\d{4}-\d{2}-\d{2}-\d{2}-\d{2}", item['date-time'] )
                if ms: 
                    dt = datetime.datetime.strptime( item['date-time'], "%Y-%m-%d-%H-%M" )
                    keep &= ( f_datetimerange[0] <= dt and dt <= f_datetimerange[1] )
                else: 
                    keep = False 

            if f_longituderange is not None:
                if item['longitude'] == float_fill_value:
                    keep = False
                else:
                    if f_longituderange[0] < f_longituderange[1]:
                        keep &= ( f_longituderange[0] <= item['longitude'] and \
                                item['longitude'] <= f_longituderange[1] )
                    else:
                        keep &= ( f_longituderange[0] <= item['longitude'] or \
                                item['longitude'] <= f_longituderange[1] )

            if f_latituderange is not None:
                if item['latitude'] == float_fill_value:
                    keep = False
                else:
                    keep &= ( f_latituderange[0] <= item['latitude'] and \
                                item['latitude'] <= f_latituderange[1] )

            if f_localtimerange is not None:
                if item['local_time'] == float_fill_value:
                    keep = False
                else:
                    if f_localtimerange[0] < f_localtimerange[1]:
                        keep &= ( f_localtimerange[0] <= item['local_time'] and \
                                item['local_time'] <= f_localtimerange[1] )
                    else:
                        keep &= ( f_localtimerange[0] <= item['local_time'] or \
                                item['local_time'] <= f_localtimerange[1] )

            if f_geometry is not None:
                if item['setting'] is None:
                    keep = False
                else:
                    keep &= ( item['setting'] and ( f_geometry == "setting" ) ) or \
                            ( ( not item['setting'] ) and ( f_geometry == "rising" ) )

            if f_availablefiletypes is not None:
                keep &= f_availablefiletypes.issubset( item.keys() )


            #  Keep or don't keep in list.

            if keep:
                keep_list.append( item )

        #  Generate new OccList based on kept items.

        return OccList( data=keep_list, s3wrapper=self._s3, version=self._version )

    def save(self, filename:str):
        """Save instance of OccList to filename in line JSON format. The OccList
        can be restored using RODatabaseClient.restore."""

        with open(filename,'w') as file:
            for item in self._data:
                file.write(json.dumps(item)+'\n')

    def info( self, param:str ) -> { list, dict }:
        '''Provides information on the following parameters: "mission", "receiver",
        "transmitter", "datetime", "longitude", "latitude", "localtime", "geometry",
        "filetype".

        mission         Return a list of the missions in the OccList instance.

        receiver        Returns a list of the receivers in the OccList instance.

        transmitter     Returns a list of the transmitters in the OccList instance.

        datetime        Returns a dictionary with the "min" and "max" ISO-format
                        date-times in the OccList instance.

        longitude       Returns a dictionary with the "min" and "max" longitudes
                        in the OccList instance.

        latitude        Returns a dictionary with the "min" and "max" latitudes
                        in the OccList instance.

        localtime       Returns a dictionary with the "min" and "max" localtimes
                        in the OccList instance.

        geometry        Returns a dictionary of the counts of "rising" and
                        "setting" occultations in the OccList instance.

        filetype        Returns a dictionary of the counts of various file types,
                        with the names of the file types as the keys and the
                        counts as their values.
        '''

        #  Show which filter key word to use based on the column they choose to show.

        option_list = []

        if param == 'datetime':
            option_list = [ item['date-time'] for item in self._data ]
            display = { 'min': min( option_list ), 'max': max( option_list ) }

        elif param in [ 'longitude', 'latitude', 'localtime' ]:
            xm = np.ma.masked_equal( [ item[param] for item in self._data ], float_fill_value )
            display = { "min":float(xm.min()), "max":float(xm.max()) }

        elif param in [ 'mission', 'receiver', 'transmitter' ]:
            display = sorted( list( { item[param] for item in self._data } ) )

        elif param == 'geometry':
            option_list = [ item['setting'] for item in self._data ]
            display = { 'nsetting':option_list.count(True), 'nrising':option_list.count(False) }

        elif param == "filetype":
            display = {}
            for item in self._data:
                for key in item.keys():
                    m = re.search( r"^(\w+)_(\w+)$", key )
                    if m:
                        if m.group(1) in valid_processing_centers[self._version] and m.group(2) in valid_file_types[self._version]:
                            if key not in display.keys():
                                display.update( { key: 0 } )
                            display[key] += 1

        return display

    def download(self, filetype:str, data_root:str=None, 
                 keep_aws_structure:bool=True, silent:bool=False ) -> list :
        """Download RO data files from AWS Registry of Open Data metadata_root of RO 
        data. 

        Download RO data of file type "filetype" from the AWS Registry of Open
        Data. The data are downloaded into the directory "data_root" as specified in 
        the defaults (created by setdefaults) data_root is specified by the keyword. 

        Arguments
        =========

        filetype            The filetype must be one of *_calibratedPhase, 
                            *_refractivityRetrieval, *_atmosphericRetrieval, where 
                            * is one of the valid contributed RO retrieval centers 
                            "ucar", "romsaf", "jpl", etc. 

        data_root              The path to the directory for downloaded RO data files. 
                            It overrides the default download path created by 
                            setdefaults. It can be a relative or absolute path. 

        keep_aws_structure  If true, create a directory hierarchy in the same way 
                            as exists in the RO metadata_root in the AWS Registry of 
                            Open Data. If false, all files are downloaded into the 
                            same directory. Note that all RO files are downloaded 
                            using the AWS hierarchy structure if data_root is not 
                            specified as an argument here. 

        silent              By setting to True, no progress bars are displayed. 
                            Progress bars are shown by default. 
        """

        if data_root is not None: 
            rootdir = data_root

        else: 
            defaults = get_defaults()
            rootdir = defaults['data_root']
            if not keep_aws_structure: 
                raise AWSgnssroutilsError( "BadArgument", 'keep_aws_structure must be ' + \
                        'true if RO data are downloaded into the default directory' )

        m = re.match( r"([a-z]+)_([a-zA-Z]+)", filetype )
        if m:
            if m.group(1) not in valid_processing_centers[self._version]:
                raise AWSgnssroutilsError( "InvalidInput", 
                        f'Invalid retrieval center "{m.group(1)}" ' + \
                        'requested; must be one of ' + \
                        ', '.join( valid_processing_centers[self._version] ) )
            elif m.group(2) not in valid_file_types[self._version]:
                raise AWSgnssroutilsError( "InvalidInput", 
                        f'Invalid file type "{m.group(2)}" ' + \
                        'requested; must be one of ' + \
                        ', '.join( valid_file_types[self._version] ) )
        else:
            raise AWSgnssroutilsError( "InvalidInput", 
                        f'You must select the "filetype" to download ' + 'as ' + \
                        ', '.join( [ f"*_{ft}" for f in valid_file_types[self._version] ] ) + \
                        ', where * is one of the processing centers ' + \
                        ', '.join( valid_processing_centers[self._version] ) )

        ro_file_list = sorted( [ item[filetype] for item in self._data if filetype in item.keys() ] )

        sTime = time.time()
        local_file_list = []

        #  Progress bar or no progress bar. 

        if silent: 
            iterator = ro_file_list
        else: 
            iterator = tqdm( ro_file_list, desc=f'Downloading {filetype}' )

        for ro_file in iterator: 

            if keep_aws_structure:
                local_path = os.path.join( rootdir, os.path.dirname(ro_file) )
            else:
                local_path = rootdir

            local_file = os.path.join( local_path, os.path.basename(ro_file) )

            #  Make the local directory path if it doesn't already exist.

            os.makedirs(local_path, exist_ok=True)

            #  Download the file if it doesn't already exist locally.

            if not os.path.exists( local_file ):
                self._s3.download( ro_file, local_file )
                ret = True
            else:
                ret = False

            local_file_list.append( local_file )

        return local_file_list

    def values( self, field:str ) -> np.ndarray :
        """Return an ndarray of values of a requested field for the data in the
        OccList. 

        Valid fields are "longitude", "latitude", "datetime", "localtime" and "time".  
        Longitudes and latitudes are in degrees; time is an ISO format time; and 
        local times are in hours."""

        if field == "longitude":
            x = np.ma.masked_equal( [ item['longitude'] for item in self._data ], float_fill_value )

        elif field == "latitude":
            x = np.ma.masked_equal( [ item['latitude'] for item in self._data ], float_fill_value )

        elif field == "localtime":
            x = np.ma.masked_equal( [ item['local_time'] for item in self._data ], float_fill_value )

        elif field == "datetime":
            x = [ item['date-time'] for item in self._data ]

        elif field == "time":
            x = [ item['time'] for item in self._data ]

        else:
            raise AWSgnssroutilsError( "InvalidArgument", "Valid fields are " + \
                    "longitude, latitude, localtime, datetime." )

        return x

    #  Magic methods.

    def __add__(self, occlist2):

        if not isinstance( occlist2, OccList ):
            raise AWSgnssroutilsError( "FaultyAddition", 
                    "Unable to concatenate; both arguments must be instances of OccList." )

        return OccList( data=self._data + occlist2._data, s3wrapper=self._s3, version=self._version )

    def __padd__(self, occlist2):

        if not isinstance( occlist2, OccList ):
            raise AWSgnssroutilsError( "FaultyAddition", 
                    "Unable to concatenate; both arguments must be instances of OccList." )

        return OccList( data=self._data + occlist2._data, s3wrapper=self._s3, version=self._version )

    def __getitem__(self,items):
        new = self._data[items]
        if isinstance( new, dict ):
            out = OccList( data=[new], s3wrapper=self._s3, version=self._version )
        else:
            out = OccList( data=new, s3wrapper=self._s3, version=self._version )
        return out

    def __repr__(self):
        return f'OccList({len(self._data)} items)'


################################################################################
#  Define the RODatabaseClient class, which creates a portal to the database
#  of radio occultation data in the AWS Registry of Open Data.
################################################################################

class RODatabaseClient:
    '''Class to initialize filter of dynamo line JSON files named by mission and day.
    An instance of this class initiates a gateway to the database of GNSS radio
    occultation data in the AWS Registry of Open Data. '''

    def __init__( self, metadata_root:str=None, version:str=None, update:bool=False ):
        '''Create an instance of RODatabaseClient. This object serves as a portal
        to the database contents of the AWS Registry of Open Data metadata_root of
        GNSS radio occultation data.

        metadata_root  If set, it is the path to the directory on the local file
                    system where the contents of the RO database are stored
                    locally. If not set, the metadata_root path is read from a 
                    defaults file created by the function setdefaults. 

        version     The version of the contents of the AWS Registry of Open Data
                    that should be accessed. The various versions that are
                    accessible can be found in s3://gnss-ro-data/dynamo/.

        update      If requested, update the contents of the local metadata_root
                    to what currently exists in the AWS metadata_root.
        '''

        #  Instantiate the s3 client in AWS region AWSregion and with unsigned certificate
        #  authentication.

        self._s3 = S3Wrapper( use_S3Client, databaseS3bucket )

        #  Check version. 

        if version is None: 
            defaults = get_defaults()
            if defaults['version'] is None: 
                raise( "InvalidVersion", 'Either specify a version when instantiating ' + \
                        'RODatabaseClient or specify it when setting defaults with setdefaults.' )
            else: 
                uversion = defaults['version']
        else: 
            uversion = version

        if uversion not in valid_versions: 
            raise AWSgnssroutilsError( "InvalidVersion", f'Version "{uversion}" is invalid; ' + \
                    'valid versions are ' + ", ".join( valid_versions ) )

        self._version = uversion

        #  Get location of local database metadata_root of previous searches. 

        if metadata_root is None: 
            defaults = get_defaults()
            self._metadata_root = defaults['metadata_root']
        else: 
            self._metadata_root = metadata_root

        os.makedirs( os.path.join( self._metadata_root, self._version ), exist_ok=True )

        #  Update the existing metadata_root if requested.

        self._update = update
        if update: 
            self.update_repo()

    def update_repo(self):
        '''This module will check for updated json files on the AWS Registry of Open Data.
        If there has been an update in files that are part of your local repo, they will be
        updated.'''

        if not os.path.exists( self._metadata_root ):
            return

        sTime = time.time()
        altzone = int( time.altzone / 3600 )     #  Correct for the time zone.
        allfiles = sorted( os.listdir( self._metadata_root ) )

        for filename in allfiles:

            linux_time = os.path.getmtime( os.path.join( self._metadata_root, filename ) )
            local_LastModified = linux_epoch + datetime.timedelta( seconds=linux_time )

            s3_info = self._s3.info( f'dynamo/{self._version}/export_subsets/{filename}' )
            s3_LastModified = s3_info['LastModified']

            if s3_LastModified > local_LastModified + datetime.timedelta(seconds=60):
                self._s3.download( s3_uri, os.path.join(self._metadata_root,self._version,filename) )

    def query(self, missions:{str,tuple,list}=None, 
              datetimerange:{tuple,list}=None, 
              silent:bool=False, **filterargs ) -> OccList:
        """Execute an query on the RO database for RO soundings. 

        This method obtains database JSON files from the AWS S3 bucket if the 
        requested files are not already available in the local metadata_root. 
        At least one of the keywords "missions" or "datetimerange" must be 
        specified. All other keywords will serve as filters on the query. 
        Return an instance of class OccList. 

        Arguments
        =========
        missions        The names of the missions to query, as defined in the 
                        documentation at github.com/gnss-ro/aws-opendata. 

        datetimerange   A two-element tuple or list specifying the time range 
                        over which to query RO metadata, both elements being 
                        ISO format datetimes.

        silent          Whether or not to run silently, generally pertaining 
                        to progress bars. 
        """

        #  Check input.

        if missions is None and datetimerange is None:
            raise AWSgnssroutilsError( "InvalidInput", 
                    f"Either 'missions' or 'datetimerange' or both must be provided." )

        #  Get listing of all JSON database files.

        file_array = self._s3.ls( f'dynamo/{self._version}/export_subsets/' )['keys']

        #  Filter by mission.

        if missions is not None:
            if isinstance( missions, str ): 
                list_missions = [ missions ]
            elif isinstance( missions, list ) or isinstance( missions, tuple ): 
                list_missions = list( missions )
            file_array = [ file for file in file_array if re.split( r"_", re.split( r"/", file )[-1] )[0] in list_missions ]

        #  Filter by date.

        if datetimerange is not None:

            try:
                dt = datetime.datetime.fromisoformat( datetimerange[0] )
                rangeStart = datetime.datetime( dt.year, dt.month, dt.day )
                dt = datetime.datetime.fromisoformat( datetimerange[1] )
                rangeEnd = datetime.datetime( dt.year, dt.month, dt.day )
            except:
                raise AWSgnssroutilsError( "FaultyDateFormat", "The datetimerange elements are not ISO format datetimes" )

            retain_array = []
            for file in file_array:
                m = re.search( r"\w+_(\d{4}-\d{2}-\d{2})\.json$", file ) 
                file_datetime = datetime.datetime.fromisoformat( m.group(1) )
                if file_datetime >= rangeStart and file_datetime <= rangeEnd:
                    retain_array.append( file )
                else: 
                    pass
            file_array = retain_array



        local_file_array = []
        os.makedirs(self._metadata_root, exist_ok=True)

        #  Progress bar? 

        if silent: 
            iterator = file_array
        else: 
            iterator = tqdm( file_array, desc="Downloading metadata" )

        for file in iterator: 
            local_path = os.path.join(self._metadata_root,self._version,os.path.basename(file))
            if not os.path.exists(local_path):
                self._s3.download( file, local_path )
            local_file_array.append(local_path)

        #  Reset file_array to local path.

        file_array = local_file_array

        #  With file array, open up and read files in to query more.

        ret_list = OccList( data=[], s3wrapper=self._s3, version=self._version )

        if silent: 
            iterator = file_array
        else: 
            iterator = tqdm( file_array, desc="Loading metadata" )

        for file in iterator:

            with open(file, 'r') as f:
                df_dict = json.loads( f.readline() )
            df = list( df_dict.values() )

            add_list = OccList( df, s3wrapper=self._s3, version=self._version ).filter( 
                    missions=missions, datetimerange=datetimerange, **filterargs )

            ret_list += add_list

        return ret_list

    def restore( self, datafile:str ) -> OccList :
        """Restore a previously saved OccList from datafile, which is a
        JSON format file."""

        if os.path.exists( datafile ):
            with open( datafile, 'r' ) as f:
                data = [ json.loads(line) for line in f.readlines() ]
        else:
            raise AWSgnssroutilsError( "FaultyData", "Argument data must be a list " + \
                    "of RO database items or a path to a previously saved OccList." )

        occlist = OccList( data=data, s3wrapper=self._s3, version=self._version )
        return occlist

    def __repr__( self ):

        output_list = []

        output_list.append( f'metadata_root="{self._metadata_root}"' )
        output_list.append( f'version="{self._version}"' )

        if self._update:
            output_list.append( "update=True" )
        else:
            output_list.append( "update=False" )

        ret = "RODatabaseClient({:})".format( ", ".join( output_list ) )
        return ret




"""rotcol.py

Function and command line executable for the rotation-collocation algorithm. 
The function bundles together the low-level routines for operationalized 
rotation-collocation. The script does the same on the Linux command line as 
well as implementing setdefaults for usernames, passwords, root directories, 
etc.

Authors: Stephen Leroy (sleroy@aer.com)
Date: April 11, 2024
"""

#  Imports. 

import argparse
import re
import json
from getpass import getpass 
from datetime import datetime
from time import time

from awsgnssroutils.database import valid_table 
from awsgnssroutils.collocation.instruments import instruments


################################################################################
#  Wrapper for executing the rotation-collocation algorithm for collocating 
#  RO soundings with scanner soundings. 
################################################################################

def execute_rotation_collocation( missions, datetimerange, ro_processing_center, 
        nadir_instrument, nadir_satellite, outputfile, nodata=False ): 
    """Execute a rotation-collocation computation. 

    Arguments
    =========

    missions        A string, tuple, or list of GNSS RO missions to consider. 

    datetimerange   A 2-element tuple or list of times, both ISO format strings, 
                    defining the time bounds for the RO soundings. 

    ro_processing_center
                    The RO processing center to be used as the source for RO 
                    retrieval data. 

    nadir_instrument
                    The name of the nadir-scanning radiance sounding instrument 
                    whose soundings are sought for collocation with RO soundings. 

    nadir_satellite
                    The name of the satellite that hosts the nadir-scanning 
                    instrument. 

    outputfile      The name of the NetCDF output file generated by the 
                    execution. 

    nodata          When true, do not extract observational data for the 
                    collocated RO and nadir-scanner soundings. 
    """

    #  Initialize return structure. 

    ret = { 'status': None, 'messages': [], 'comments': [] }

    #  Relevant imports. 

    from awsgnssroutils.database import RODatabaseClient
    from awsgnssroutils.collocation.core.spacetrack import Spacetrack
    from awsgnssroutils.collocation.core.spacetrack import checkdefaults as spacetrack_checkdefaults
    from awsgnssroutils.collocation.core.rotation_collocation import rotation_collocation

    #  Initialize RO database client and access to Space-Track data. 

    ret = spacetrack_checkdefaults()
    if ret['status'] == "fail": 
        print( "Error: \n" + "\n".join( ret['comments'] ) )
        return

    rodb = RODatabaseClient()
    strack = Spacetrack()

    #  Instantiate the nadir-scanner instrument. 

    from awsgnssroutils.collocation.core.nasa_earthdata import Satellites as NASA_Satellites
    from awsgnssroutils.collocation.core.eumetsat import metop_satellites as EUMETSAT_Satellites

    if nadir_satellite in NASA_Satellites.keys(): 

        from awsgnssroutils.collocation.core.nasa_earthdata import checkdefaults, NASAEarthdata
        ret = checkdefaults()
        if ret['status'] == "fail": 
            print( "Error: \n" + "\n".join( ret['comments'] ) )
            return

        data_center = "NASA Earthdata" 
        nasa_earthdata_access = NASAEarthdata()
        inst = instruments[nadir_instrument]['class']( nadir_satellite, nasa_earthdata_access, spacetrack=strack )

    elif nadir_satellite in EUMETSAT_Satellites: 

        from awsgnssroutils.collocation.core.eumetsat import checkdefaults, EUMETSATDataStore
        ret = checkdefaults()
        if ret['status'] == "fail": 
            print( "Error: \n" + "\n".join( ret['comments'] ) )
            return

        data_center = "EUMETSAT Data Store"
        eumetsat_data_store = EUMETSATDataStore()
        inst = instruments[nadir_instrument]['class']( nadir_satellite, eumetsat_data_store, spacetrack=strack )

    #  Collocation tolerances. 

    time_tolerance = 600                           # 10 min/600 sec
    spatial_tolerance = 150.0e3                    # m

    #  Get occultation geolocations. 

    print( "Querying occultation database" )

    tbegin = time()
    occs = rodb.query( missions=missions, datetimerange=[ dt.isoformat() for dt in datetimerange ], 
                availablefiletypes=f'{ro_processing_center}_refractivityRetrieval', silent=True )
    tend = time()

    if occs.size == 0: 
        ret['status'] = "fail"
        ret['messages'].append( "NoOccultationData" )
        ret['comments'].append( "No occultation data found for missions and time range" )
        return ret

    print( "  - number found = {:}".format( occs.size ) )
    print( "  - elapsed time = {:10.3f} s".format( tend-tbegin ) )

    #  Exercise rotation-collocation. 

    print( "Executing rotation-collocation" )

    tbegin = time()
    ret_rotation = rotation_collocation( inst, occs, time_tolerance, spatial_tolerance, 2 )
    tend = time()

    ret['messages'] += ret_rotation['messages']
    ret['comments'] += ret_rotation['comments']

    if ret_rotation['status'] == "fail": 
        ret['status'] = "fail"
        print()
        return ret

    collocations_rotation = ret_rotation['data']

    print( "  - number found = {:}".format( len( collocations_rotation ) ) )
    print( "  - elapsed time = {:10.3f} s".format( tend-tbegin ) )

    if not nodata: 

        #  Populate instrument data. 

        print( f'Downloading instrument data from {data_center}' )
        tbegin = time()
        ret_populate = inst.populate( datetimerange )
        tend = time()

        print( "  - elapsed time = {:10.3f} s".format( tend-tbegin ) )

        #  Extract data. 

        print( "Extracting collocation data" )

        tbegin = time()
        for collocation in collocations_rotation: 
            occid = collocation.get_data( ro_processing_center )
        tend = time()

        print( "  - elapsed time = {:10.3f} s".format( tend-tbegin ) )

    #  Save to output file. 

    tbegin = time()
    print( f"Writing to output file {outputfile}" )
    collocations_rotation.write_to_netcdf( outputfile )
    tend = time()

    print( "  - elapsed time = {:10.3f} s".format( tend-tbegin ) )

    #  Done. 

    ret['status'] = "success" 

    return ret


################################################################################
#  Command line parser and execution of the above function. 
################################################################################

def main(): 

    #  Choices and defaults. 

    version = "v1.1"
    output_default = "output.nc"
    ro_processing_center_default = "ucar"

    all_valid_missions = sorted( list( set( [ item['mission'] for item in valid_table 
            if item['version']==version and item['filetype'] == "refractivityRetrieval" ] ) ) )

    all_valid_centers = sorted( list( set( [ item['center'] for item in valid_table 
            if item['version']==version and item['filetype'] == "refractivityRetrieval" ] ) ) )

    all_valid_instruments = sorted( list( instruments.keys() ) )

    all_valid_satellites = []
    for key, val in instruments.items(): 
        all_valid_satellites += val['valid_satellites']
    all_valid_satellites = sorted( list( set( all_valid_satellites ) ) )

    #  Define the parent (root) parser. 

    parser = argparse.ArgumentParser( prog='rotcol', add_help=True, 
            description="""This program performs a collocation-finding calculation by between 
            GNSS radio occultation data and passive nadir-scanner sounding data by implementing 
            the rotation-collocation algorithm, generating NetCDF output containing the 
            data associated with the collocations. The user must set various defaults 
            (setdefaults) before executing (execute) a rotation-collocation computation.""", 
            epilog=" " )

    subparsers = parser.add_subparsers( dest="command" )

    #  Define the setdefaults parser. 

    setdefaults_parser = subparsers.add_parser( "setdefaults", 
            help="""Set default paths for data and metadata, and define login fields 
            (username, password, etc.) for access to various services. The services are 
            "awsro", "eumetsat", "earthdata", and "spacetrack".""" )

    setdefaults_parser.add_argument( "service", type=str, 
            help=
            """The service for which to set defaults. 

            (1) "awsro" refers to the AWS Registry of Open Data repository for GNSS radio 
            occultation data. If requested, the user should provide the dataroot and the 
            metadataroot paths where the data and RO metadata should be stored on the local 
            file system.

            (2) "eumetsat" refers to the EUMETSAT Data Store. If requested, the user 
            should consider entering the dataroot where EUMETSAT Data Store should be stored 
            on the local file system; also, the ConsumerKey and SecretKey that provides online 
            access to the EUMETSAT Data Store, which can be obtained from the EUMETSAT Data 
            Store website. 

            (3) "earthdata" refers to the NASA Earthdata portal to the NASA DAACs. If 
            requested, the user should consider entering the dataroot where Earthdata data 
            should be stored on the local file system; also the username and password for the 
            user's account in NASA Earthdata.

            (4) "spacetrack" refers to the Space-Track portal for orbital TLE data. If 
            requested, the user should consider entering the dataroot where Space-Track TLE 
            data should be stored on the local file system; also, the username and password 
            for the user's Space-Track account.""" ) 

    setdefaults_parser.add_argument( "--dataroot", dest="dataroot", default=None, type=str, required=False, 
            help="""Path on the local file system where data should be downloaded and stored 
            for future reference (and efficiency).""" )

    setdefaults_parser.add_argument( "--metadataroot", dest="metadataroot", default=None, type=str, required=False, 
            help="""Path on the local file system where metadata should be downloaded and stored 
            for future reference (and efficiency).""" )

    setdefaults_parser.add_argument( "--username", dest="username", default=None, type=str, required=False, 
            help="""Username for the Earthdata or Space-Track account; user will be prompted for password.""" )

    setdefaults_parser.add_argument( "--consumer-key", dest="consumer_key", default=None, type=str, required=False, 
            help="""ConsumerKey as provided by the EUMETSAT Data Store.""" )

    setdefaults_parser.add_argument( "--consumer-secret", dest="consumer_secret", default=None, type=str, required=False, 
            help="""ConsumerSecret as provided by the EUMETSAT Data Store.""" )

    #  Define the execution parser. 

    collocation_parser = subparsers.add_parser( "execute", 
            help="""Execute a rotation-collocation calculation to find 
            collocations between GNSS radio occultation soundings and a selected 
            nadir-scanning instrument. In this the user must list the RO missions 
            from which to draw data, what nadir-scanning instrument on what satellite 
            to search for collocated sounder data, and over what time period 
            collocations are being sought.""" )

    missions_default = "cosmic1 cosmic2 metop"
    collocation_parser.add_argument( "missions", type=str, default=missions_default, 
            help="""A list of GNSS radio occultation missions to draw from, separated by white space. Choices 
            are """ + ", ".join( all_valid_missions ) + f'.  The default is "{missions_default}".' )

    collocation_parser.add_argument( "timerange", type=str, 
            help="""Two ISO-format UTC datetimes defining the time interval over which to search for 
            collocations, separated by spaces""" )

    collocation_parser.add_argument( "nadir_instrument", type=str, 
            help="""Name of the nadir-scanning instrument. Choices are """ + \
                    ', '.join( all_valid_instruments ) + "." )

    collocation_parser.add_argument( "nadir_satellite", type=str, 
            help="""Name of the satellite hosting the nadir-scanning instrument. Choices are """ + \
                    ', '.join( all_valid_satellites ) + "." )

    collocation_parser.add_argument( "--nodata", default=False, action="store_true", 
            help="""Write geolocation information on collocation, but do not extract any RO profile or 
                    nadir-scanning instrument data. The default is to extract and write sounding data 
                    into the output file.""" )

    collocation_parser.add_argument( "--ro-center", dest="ro_processing_center", type=str, 
            choices=all_valid_centers, default=ro_processing_center_default, 
            help='Name of the RO processing center from which to take data; by default ' + \
                    f'"{ro_processing_center_default}".' )

    collocation_parser.add_argument( "--output", "-o", type=str, default=output_default, 
            help=f'Output file name for collocation data; by default "{output_default}".' )

    #  Parse arguments. 

    args = parser.parse_args()

    if args.command == "setdefaults": 

        #  Set defaults. 

        if args.service == "awsro": 

            from awsgnssroutils.database import setdefaults

            kwargs = { 'version': "v1.1" }
            if args.dataroot is not None: 
                kwargs.update( { 'data_root': args.dataroot } )
            if args.metadataroot is not None: 
                kwargs.update( { 'metadata_root': args.dataroot } )

            print( "Updating AWS RO defaults: " + json.dumps( kwargs ) )
            setdefaults( **kwargs )

        elif args.service == "eumetsat": 

            from awsgnssroutils.collocation.core.eumetsat import setdefaults

            kwargs = {}
            if args.dataroot is not None: 
                kwargs.update( { 'root_path': args.dataroot } )
            if args.consumer_key is not None and args.consumer_secret is not None: 
                kwargs.update( { 'eumetsattokens': ( args.consumer_key, args.consumer_secret ) } )

            print( "Updating EUMETSAT defaults: " + json.dumps( kwargs ) )
            ret = setdefaults( **kwargs )
            if ret['status'] == "fail": 
                print( "Error: \n" + "\n".join( ret['comments'] ) )

        elif args.service == "earthdata": 

            from awsgnssroutils.collocation.core.nasa_earthdata import setdefaults

            kwargs = {}
            if args.dataroot is not None: 
                kwargs.update( { 'root_path': args.dataroot } )
            if args.username is not None: 
                password = getpass()
                kwargs.update( { 'earthdatalogin': ( args.username, password ) } )

            print( "Updating NASA Earthdata defaults: " + json.dumps( kwargs ) )
            ret = setdefaults( **kwargs )
            if ret['status'] == "fail": 
                print( "Error: \n" + "\n".join( ret['comments'] ) )

        elif args.service == "spacetrack": 

            from awsgnssroutils.collocation.core.spacetrack import setdefaults

            kwargs = {}
            if args.dataroot is not None: 
                kwargs.update( { 'root_path': args.dataroot } )
            if args.username is not None: 
                password = getpass()
                kwargs.update( { 'spacetracklogin': ( args.username, password ) } )

            print( "Updating Space-Track defaults: " + json.dumps( kwargs ) )
            ret = setdefaults( **kwargs )
            if ret['status'] == "fail": 
                print( "Error: \n" + "\n".join( ret['comments'] ) )

        else: 

            print( f'Invalid service: "{args.service}"' )

    elif args.command == "execute": 

        #  Execute rotation-collocation algorithm for collocation finding. 

        #  Get arguments for rotation-collocation: missions, timerange, nadir_satellite, 
        #  nadir_instrument, ro_processing_center. 

        missions = re.split( "\s+", args.missions )

        ss = re.split( "\s+", args.timerange )
        timerange = ( datetime.fromisoformat( ss[0] ), datetime.fromisoformat( ss[1] ) )

        nadir_satellite = str( args.nadir_satellite )
        nadir_instrument = str( args.nadir_instrument )
        ro_processing_center = str( args.ro_processing_center )
        output_file = str( args.output )

        #  Check for valid satellite-instrument combination. 

        if nadir_instrument not in instruments.keys(): 

            print( f'Instrument "{nadir_instrument}" is unrecognized. Valid satellite ' + \
                    'instruments are ' + ", ".join( sorted( list( instruments.keys() ) ) ) + '.' )

        elif nadir_satellite not in instruments[nadir_instrument]['valid_satellites']: 

            print( f'Satellite {nadir_satellite} has no instrument {nadir_instrument}, ' + \
                'or at least it is not registered in this package. The instrument ' + \
                f'{nadir_instrument} has been incorporated for the following satellites: ' + \
                ", ".join( instruments[nadir_instrument]['valid_satellites'] ) + "." )

        else: 

            ret = execute_rotation_collocation( missions, timerange, ro_processing_center, 
                    nadir_instrument, nadir_satellite, output_file, nodata=args.nodata )

            if ret['status'] == "fail": 
                print( 'Unsuccessful. {:}: {:}'.format( ret['messages'][-1], ret['comments'][-1] ) )
            else: 
                print( 'Successful completion.' )

    else: 
        print( 'No command provided. Valid commands are "setdefaults" and "execute".' )
            
    return


################################################################################
#  Enable execution of this script from the command line in development mode. 
################################################################################

if __name__ == "__main__": 
    main()
    pass




"""
The `rotation_collocation` module contains the core implementation 
of the rotation-collocation methods detailed in Meredith, Leroy,
Halperin, and Cahoy 2023. It also includes code to extract 
nadir-scanner soundings associated with collocations.

Last updated: 9/8/2023 by Alex Meredith
"""

import bisect
import numpy as np
from sgp4.api import Satrec
from .nadir_satellite import NadirSatelliteInstrument
from .awsro import get_occ_times
from .collocation import Collocation, CollocationList
from .constants_and_utils import km_to_degree, constrain_to_pi_range, constrain_angle_pair


#  Exception handling. 

class Error( Exception ): 
    pass

class rotationCollocationError( Error ): 
    def __init__( self, message, comment ): 
        self.message = message
        self.comment = comment



def rotation_collocation( nadir_sat, occs, time_tolerance, spatial_tolerance, 
                         nsuboccs=2 ): 
    """This function performs collocation-finding using the rotation-collocation
    method. If nsuboccs > 2, this is equivalent to the rotation-collocation method
    with sub-occultations. If nsuboccs = 2, this is equivalent to the linearized
    rotation-collocation method.

    Arguments
    -----------
        nadir_sat: subclass of NadirSatelliteInstrument
            Nadir-scanning satellite object
        occs: OccList
            A list of occultations (and their metadata)
        time_tolerance: float
            Time tolerance, in seconds
        spatial_tolerance: float
            Spatial tolerance, in m
        nsuboccs: int
            Number of sub-occultations (use 2 for linearized)

    Returns
    ---------
        A dictionary with keyword value pairs as follows: 

        status -> "success" or "fail"
        messages -> a list of mnemonic messages regarding performance
        comments -> a list of commentary messages regarding performance
        data -> an instance of CollocationList containing found collocations
    """

    #  Initialize return dictionary. 

    ret = { 'status': None, 'messages': [], 'comments': [], 'data': None }

    #  Check input. 

    if not isinstance( nadir_sat, NadirSatelliteInstrument ): 
        ret['messages'].append( "InvalidArgument" )
        ret['comments'].append( "nadir_sat must be a subclass of NadirSatelliteInstrument" )
        ret['status'] = "fail"
        return ret

    if nsuboccs < 2:
        ret['messages'].append( "InvalidArgument" )
        ret['comments'].append( 
            """Rotation-collocation method needs at least 2 
            points for interpolation. Use nsuboccs=2 for the linearized 
            rotation-collocation method and use nsuboccs>2 for the 
            rotation-collocation method with sub-occultations""" )
        ret['status'] = "fail"
        return ret

    ret_rotate = nadir_sat.rotate_to_nadir_frame( occs, time_tolerance, nsuboccs )
    if ret_rotate['status'] == "fail": 
        ret['status'] = "fail"
        ret['messages'] += ret_rotate['messages']
        ret['comments'] += ret_rotate['comments']
        return ret
    nadir_frame_positions = ret_rotate['data']

    dt = 2 * time_tolerance/(nsuboccs-1)

    colocs = []

    #  Get satellite period (assume it doesn't change much over the day)

    time = get_occ_times( occs[0] )[0]
    resp = nadir_sat.get_current_tle( time )

    if resp['status'] == "fail": 
        ret['status'] == "fail"
        ret['messages'] += resp['messages']
        ret['comments'] += resp['comments']
        return ret
    else: 
        tle = resp['data']

    sat = Satrec.twoline2rv( *tle )
    sat_period = 2*np.pi/sat.nm * 60        #  Convert to seconds from minutes

    for iocc in range(occs.size): 

        occ = occs[iocc]
        time = get_occ_times( occ )[0]
        latitude = np.deg2rad( occ.values("latitude")[0] )
        longitude = np.deg2rad( occ.values("longitude")[0] )

        #  Get max scan angle (depends on latitude + nadir-scanner altitude)

        max_scan_angle = nadir_sat.calculate_ds_with_sgp4( time, latitude, spatial_tolerance, sat )

        for isubocc in range(1,nsuboccs):

            sub_occ_pos_prev = nadir_frame_positions[iocc, isubocc-1, :, 0]
            sub_occ_time_prev = time + (isubocc-1-(nsuboccs-1)/2)*dt
            sub_occ_pos = nadir_frame_positions[iocc, isubocc, :, 0]
            sub_occ_time = time + (isubocc-(nsuboccs-1)/2)*dt

            x_prev = sub_occ_pos_prev[0]
            y_prev = sub_occ_pos_prev[1]
            z_prev = sub_occ_pos_prev[2]
            x, y, z = sub_occ_pos[0], sub_occ_pos[1], sub_occ_pos[2]

            #  Number of expected orbits between sub-occultations. This is
            #  important for making sure we don't miss collocations when
            #  the time tolerance is quite long.

            num_revs = ( sub_occ_time - sub_occ_time_prev ) / sat_period
            ret_compare = compare_points_in_nadir_frame(x_prev, y_prev, z_prev,
                    x, y, z, max_scan_angle, num_revs, spatial_tolerance)

            if ret_compare is None: 
                continue

            frac = ( ret_compare['estimated_scan_angle'] - ret_compare['start_scan_angle'] ) \
                        / ( ret_compare['end_scan_angle'] - ret_compare['start_scan_angle'] )

            est_coloc_time = sub_occ_time_prev
            est_coloc_time += frac * ( sub_occ_time - sub_occ_time_prev )

            max_scan_angle = nadir_sat.calculate_ds_with_sgp4( est_coloc_time, 
                    latitude, spatial_tolerance, sat )

            # Double check to make sure max scan angle is exactly right
            # at estimated collocation time

            if ret_compare['estimated_scan_angle'] <= max_scan_angle:
                coloc = Collocation( occ, nadir_sat, 
                                    longitude=np.rad2deg(longitude), 
                                    latitude=np.rad2deg(latitude), 
                                    scan_angle=np.rad2deg(ret_compare['estimated_scan_angle']), 
                                    time=est_coloc_time )
                colocs.append( coloc )
                break

    ret['status'] = "success"
    ret['data'] = CollocationList( colocs )

    return ret


def compare_points_in_nadir_frame(x_prev, y_prev, z_prev, x_j, y_j, z_j,
                                  max_scan_angle, num_revs, spatial_tolerance):
    """
    This function checks if two points in the nadir satellite frame are
    collocated. It is a helper used by the rotation-collocation method.

    Arguments
    ----------
        x_prev : float
            x-coordinate of start point in nadir satellite frame
        y_prev : float
            y-coordinate of start point in nadir satellite frame
        z_prev : float
            z-coordinate of start point in nadir satellite frame
        x_j : float
            x-coordinate of end point in nadir satellite frame
        y_j : float
            y-coordinate of end point in nadir satellite frame
        z_j : float
            z-coordinate of end point in nadir satellite frame
        max_scan_angle : float
            maximum scan angle of the nadir satellite
        num_revs: float
            fractional number of orbits completed between t_prev and t_j
        spatial_tolerance: float
            spatial tolerance for colocation in m

    Returns
    ----------
        None if no collocation is found. If a collocation is found, then a 
        dictionary is returned with the following keyword-value pairs: 

        estimated_scan_angle -> float
            Estimated scan angle (radians) of the collocated sounding
        start_scan_angle -> float
            Scan angle (radians) of the sub-occultation preceding the collocation
        end_scan_angle -> float
            Scan angle (radians) of the sub-occultation immediately succeding the 
            collocation. 
    """

    is_coloc = False

    # Find start and end arglat
    start_delta_arglat = np.arctan2(y_prev, x_prev)
    end_delta_arglat = np.arctan2(y_j, x_j)

    # Shift delta arglats such that end < start and |end-start| < 2pi
    (end_delta_arglat,
     start_delta_arglat) = constrain_angle_pair(end_delta_arglat,
                                                start_delta_arglat,
                                                num_full_revs=num_revs)
    min_delta_arglat = end_delta_arglat
    max_delta_arglat = start_delta_arglat

    # Find start and end scan angle
    start_scan_angle = np.arcsin(z_prev)
    end_scan_angle = np.arcsin(z_j)

    # Shift scan angles to be within -pi and pi
    start_scan_angle = constrain_to_pi_range(start_scan_angle)
    end_scan_angle = constrain_to_pi_range(end_scan_angle)

    # Find scan angle in terms of arglat (linear fit)
    rise = end_scan_angle-start_scan_angle
    run = end_delta_arglat-start_delta_arglat
    slope = rise/run

    # Need to check for 2pi crossing if num_full_revs is big enough, etc.
    i = 0
    zero_crossing = 0
    while i*2*np.pi < start_delta_arglat:
        y_int = end_scan_angle - slope*end_delta_arglat
        zero_crossing = y_int + i*slope*2*np.pi

        # Check for zero crossing within scan range
        if min_delta_arglat < 0 and max_delta_arglat > 0:
            if np.abs(zero_crossing) < max_scan_angle:
                is_coloc = True
                break
        i += 1

    # Check for near-zero delta arglat within scan range
    spatial_tol_rad = np.deg2rad(spatial_tolerance*1.0e-3*km_to_degree)
    if np.cos(start_delta_arglat) > np.cos(spatial_tol_rad):
        if np.abs(start_scan_angle) < max_scan_angle:
            is_coloc = True
            zero_crossing = start_scan_angle
    if np.cos(end_delta_arglat) > np.cos(spatial_tol_rad):
        if np.abs(end_scan_angle) < max_scan_angle:
            is_coloc = True
            zero_crossing = end_scan_angle

    if is_coloc: 
        ret = { 'estimated_scan_angle': zero_crossing, 
                'start_scan_angle': start_scan_angle, 
                'end_scan_angle': end_scan_angle }
    else: 
        ret = None

    return ret




"""
Module: nadir_satellite.py
Authors: Alex Meredith, Stephen Leroy
Contact: sleroy@aer.com
Date: April 10, 2024

This module contains two classes that define nadir-scanning satellite 
instruments and provide a context for storing geolocation metadata from 
their data files. 

NadirSatelliteInstrument
========================
Provides information on the details of how a nadir-scanning satellite 
instrument goes about scanning. It also includes a suite of methods 
that can be used in applying the rotation-collocation algorithm. Those 
methods are...

    - get_geolocations: Gets geolocation metadata for a specified 
            time range; it populates an instance of ScanMetadata
    - get_current_tle: Gets two-line element for the host satellite 
            that is nearest to a specified time; it accesses 
            Spacetrack
    - calculate_ds_from_xi: Calculate approximate maximum scan width 
            given a geodetic latitude; it only computes the local 
            radius of the Earth to estimate the scan width ds 
    - calculate_ds_with_sgp4: Calculate precise maximum scan width
            given time, latitude, etc., based on SGP4 orbit 
            propgation 
    - get_inertial_lat_lon: Calculate the longitude-latitude 
            coordinates of a set of occultations in ECI coordinates. 
    - rotate_to_nadir_frame: Rotate (sub-)occultations into the 
            reference frame of a nadir scanner's scan pattern; it is 
            the heart of the rotation-collocation algorithm. 

ScanMetadata
============
Stores geolocation information and metadata useful for retrieving 
radiance data for the satellite instrument. It is a callable object, 
retrieving radiance data corresponding to a scan index and a 
footprint index for the geolocation data stored in the object. 

In order for the object to retrieve nadir-scanner radiance data, 
it must connect to a "get_data" method to be provided by the 
satellite instrument object that inherits this class. 
"""

import erfa
import numpy as np
from abc import ABC, abstractmethod
from sgp4.ext import rv2coe
from sgp4.api import Satrec
from astropy.coordinates.builtin_frames.utils import get_polar_motion

from .awsro import get_occ_times
from .constants_and_utils import sec_to_sidereal_day, calculate_km_to_degree, mu, calculate_radius_of_earth
from .spacetrack import Spacetrack

#  Exception handling. 

class Error( Exception ): 
    pass

class nadirSatelliteError( Error ): 
    def __init__( self, message, comments ): 
        self.message = message
        self.comments = comments 


class NadirSatelliteInstrument(ABC):
    """Generalized nadir-scanning satellite object. This class contains methods
    key to the function of the rotation-collocation method -- a method to
    rotate points into the coordinate frame given by the nadir-scanner
    satellite, and several helper methods for this coordinate frame
    transformation.

    Arguments
    ------------
        name: string
            Name of satellite (ex. NOAA-20)
        max_scan_angle: float
            Nadir-scanning instrument maximum scan angle [degrees]
        time_between_scans: float
            Time taken to complete one cross-track scan [seconds]
        scan_points_per_line: int
            Number of scan points in each cross-track scan
        scan_angle_spacing: float
            Angle between scan points in each cross-track scan [degrees]
        spacetrack: spacetrack.Spacetrack
            Portal to Spacetrack TLE data on the local file system

    Attributes
    ------------
        satellite_name: string
            Name of satellite (ex. NOAA-20)
        spacetrack_satellite: instance of spacetrack.SpacetrackSatellite
            An instance of spacetrack.SpacetrackSatellite for access to TLEs
        xi: float
            Nadir-scanning instrument maximum scan angle [radians]
        time_between_scans: float
            Time between cross-track scans [seconds]
        scan_points_per_line: int
            Number of individual footprints per cross-track scan
        max_scan_angle: float
            Maximum scan angle [radians]
        scan_angle_spacing: float
            Angle between scan footprints [radians]"""

    def __init__(self, satellite_name, max_scan_angle, 
                 time_between_scans, scan_points_per_line, scan_angle_spacing, 
                 spacetrack=None):
        """
        Constructor for NadirSatelliteInstrument.
        """
        self.satellite_name = satellite_name
        self.xi = np.deg2rad(max_scan_angle)

        self.time_between_scans = time_between_scans
        self.scan_points_per_line = scan_points_per_line
        self.scan_angle_spacing = np.deg2rad(scan_angle_spacing)
        self.max_scan_angle = np.deg2rad(max_scan_angle)

        if isinstance( spacetrack, Spacetrack ): 
            self.spacetrack_satellite = spacetrack.satellite( satellite_name )
        else: 
            self.spacetrack_satellite = None

    #====================================================================
    #  Abstract methods. These must be defined by the child class that 
    #  inherits this base class. 
    #====================================================================

    @abstractmethod
    def get_geolocations( self, timerange ):
        """This method retrieves the geolocations of nadir-scanner soundings 
        that fall within a user-prescribed timerange. It returns an instance 
        of class ScanMetadata. 

        Arguments
        ---------
            timerange: tuple/list of timestandards.Time objects
                This defines the time range of nadir-scanner geolocations 
                to be retrieved. 

        Returns
        ---------
            geolocations: ScanMetadata
                An object containing information on longitudes, latitudes, and 
                and times of nadir-scanner soundings. It also contains pointers 
                to data files and a method to retrieve observational (radiance) 
                data from those files."""

    @abstractmethod
    def get_data( self, file, scan_index, footprint_index, **kwargs ): 
        """A function which returns nadir-scan satellite data for requested scan and
        footprint indices within the file *file*. The function itself must be a
        Method to fetch data for a nadir-scan satellite instrument corresponding to
        a scalar integer indicating the scan number [0:nscans] and a scalar integer
        indicating the footprint number [0:nfootprints].

        It should return a data dictionary according taken from file *file* and the
        data location within the file should correspond to scans
        data[scan_index,footprint_index].

        Optional arguments are passed to get_data as the dictionary kwargs."""


    #====================================================================
    #  Defined methods. Mostly used in the rotation-collocation 
    #  algorithm. 
    #====================================================================

    def get_current_tle(self, occultation_time):
        """This method gets the current TLE for the satellite at a given time.
        It should be implemented by child classes appropriately, depending
        on how TLEs are stored.

        Parameters
        ------------
            occultation_time: timestandards.Time
                Time of RO sounding 

        Returns
        ---------
            tle_1: string
                first line of TLE
            tle_2: string
                second line of TLE"""

        ret = self.spacetrack_satellite.nearest( occultation_time )
        return ret

    def calculate_ds_with_xi(self, lat_geodetic, alt ):
        """This method calculates the latitude-dependent scan angle for use
        in the rotation method, with varying max scan distance. This
        method is used for debugging/visualizing only.

        Arguments
        -----------
            lat_geodetic: float
                Geodetic latitude [radians]
            alt: float
                Altitude [m], default to satellite semi-major axis

        Returns
        ---------
            ds: float
                Latitude-dependent scan angle [radians]"""

        if alt is None:
            alt = self.a
        r = calculate_radius_of_earth(lat_geodetic)
        ds = np.arcsin((alt/r)*np.sin(self.xi)) - self.xi

        return ds

    def calculate_ds_with_sgp4(self, sounding_time, sounding_lat, spatial_tolerance, sgp4_sat=None):
        """This method gets the max scan angle for a nadir-scanning satellite
        very precisely by propagating the satellite position with SGP4, getting
        nadir-scanner altitude and latitude at the necessary time, and then
        calculating the scan angle from there.

        Arguments
        ------------
            sounding_time: timestandards.Time
                time to calculate max scan angle
            sounding_lat: float
                sounding latitude for calculation, in radians
            spatial_tolerance: float
                spatial tolerance, for soundings just past true range, in m
            sgp4_sat: sgp4.wrapper.Satrec (optional)
                sgp4 satellite object with TLE already loaded, for efficiency

        Returns
        ----------
            max_scan_angle: float
                max scan angle in radians"""

        jd = sounding_time.juliandate()
        if sgp4_sat is None:
            tle = self.get_current_tle( sounding_time )
            sgp4_sat = Satrec.twoline2rv( *tle )

        e, r, v = sgp4_sat.sgp4(jd, 0)
        r_norm = (r[0]**2 + r[1]**2 + r[2]**2)**0.5
        nadir_scanner_lat = np.arcsin(r[2]/r_norm)

        ds = self.calculate_ds_with_xi(nadir_scanner_lat, r_norm)
        km_to_degree = calculate_km_to_degree(sounding_lat)
        max_scan_angle = abs(ds) + np.deg2rad(spatial_tolerance*1.0e-3*km_to_degree)

        return max_scan_angle

    def get_inertial_lat_lon(self, jd, polarmat, lon, lat):
        """Method to find the inertial latitude and longitude of an occultation.

        Arguments 
        -----------
            jd: float
                Julian date representing occultation time
            polarmat: np array
                Matrix representing polar motion
            lon: float
                Longitude of occultation [radians]
            lat: float
                Latitude of occultation [radians]
        Returns
        --------
            lon_inert: float
                Inertial longitude of occultation [radians]
            lat_inert: float
                Inertial latitude of occultation [radians]"""

        # Convert to position vector in ECEF coordinates

        pos_ITRS = np.array([[np.cos(lon)*np.cos(lat)],
                             [np.sin(lon)*np.cos(lat)],
                             [np.sin(lat)]])

        # Convert to TEME

        gst = erfa.gmst82(jd, 0)
        teme_to_itrs_mat = erfa.c2tcio(np.eye(3), gst, polarmat)
        occ_TEME = teme_to_itrs_mat.T@pos_ITRS

        # Find inertial latitude and longitude from TEME

        lon_inert = np.arctan2(occ_TEME[1][0], occ_TEME[0][0])
        lat_inert = np.arcsin(occ_TEME[2][0])

        return lon_inert, lat_inert

    def rotate_to_nadir_frame( self, occs, time_tolerance, nsuboccs ):
        """Method to get the position in the coordinate frame given by the
        nadir sounder for a list of occultations. Used by the
        rotation-collocation method.

        Arguments
        -----------
            occs: OccList
                A set of occultations given as an OccList. 
            time_tolerance: float
                Time matchup tolerance, in seconds. The two furthest-
                apart sub-occultations will be separated by
                2*time_tolerance seconds.
            nsuboccs: int
                Number of sub-occultations

        Returns
        -----------
            nadir_positions: np.ndarray (m x N x 3 x 1)
                Position rotated into the nadir-scanner frame for each
                sub-occultation. nadir_positions[m][n] is the position
                vector in the nadir-scanner frame for the nth
                sub-occultation of the mth occultation."""

        ret = { 'status': None, 'messages': [], 'comments': [], 'data': None }

        #  Get spacing between sub-occultations in sec

        if nsuboccs == 1:
            subocc_spacing = 0
        elif nsuboccs > 1:
            subocc_spacing = 2 * time_tolerance / (nsuboccs-1)

        nadir_positions = np.zeros( ( occs.size, nsuboccs, 3, 1 ), dtype=np.float64 )

        s, tline = "", ""
        polarmat = np.zeros((3, 3))

        for iocc in range(occs.size): 

            occ = occs[iocc]
            lon = np.deg2rad( occ.values("longitude")[0] )
            lat = np.deg2rad( occ.values("latitude")[0] )
            time = get_occ_times( occ )[0]

            # Get polar motion and TLE (only once per day; this
            # doesn't need to be done for each occultation)

            if iocc == 0:
                xp, yp = get_polar_motion( time.juliandate() )
                polarmat = erfa.pom00( xp, yp, 0 )
                ret_tle = self.get_current_tle( time )
                ret['messages'] += ret_tle['messages']
                ret['comments'] += ret_tle['comments']
                if ret_tle['status'] == "fail": 
                    ret['status'] = "fail"
                    return ret
                tle = ret_tle['data']
                sat = Satrec.twoline2rv( *tle )

            # Find inertial longitude & latitude for occultation

            lon_inert, lat_inert = self.get_inertial_lat_lon( time.juliandate(), polarmat, lon, lat)
            sidereal_spin_per_sec = 2*np.pi * sec_to_sidereal_day

            # Loop over sub-occultations

            for isubocc in range(nsuboccs):

                # Find sub-occultation time and longitude (we assume the
                # atmosphere rotates with the Earth, so the longitude
                # changes with Earth's rotation, which can be approximated
                # by sidereal spin). 

                subocc_offset = (isubocc-((nsuboccs-1)/2)) * subocc_spacing
                subocc_time = time + subocc_offset
                subocc_lon = lon_inert + sidereal_spin_per_sec * subocc_offset

                # Use sgp4 to get orbital position and orbital elements. 

                jd = subocc_time.juliandate()
                r, v = sat.sgp4(jd, 0)[1:3]
                incl, raan_lon, argp, nu = rv2coe(r, v, mu)[3:7]
                arglat = argp + nu
                lon_diff = subocc_lon - raan_lon

                # Construct matrix to rotate into nadir-scanner frame. 

                arglat_matrix = np.array([[np.cos(arglat), np.sin(arglat), 0],
                                          [-np.sin(arglat), np.cos(arglat), 0],
                                          [0, 0, 1]])

                incl_matrix = np.array([[1, 0, 0],
                                        [0, np.cos(incl), np.sin(incl)],
                                        [0, -np.sin(incl), np.cos(incl)]])

                raan_matrix = np.array([[np.cos(lon_diff)*np.cos(lat_inert)],
                                        [np.sin(lon_diff)*np.cos(lat_inert)],
                                        [np.sin(lat_inert)]])

                pos_subocc = arglat_matrix @ incl_matrix @ raan_matrix

                nadir_positions[iocc, isubocc, :, :] = pos_subocc

        #  Done. 

        ret['status'] = "success"
        ret['data'] = nadir_positions

        return ret



class ScanMetadata(): 
    """This class defines the geolocations and data pointer information for a set of 
    scans from a nadir-scan satellite. The various elements are 

    'get_data': NadirSatelliteInstrument.get_data -> Pointer to the get_data 
            method in the NadirSatelliteInstrument *child* class. 

    'longitudes': numpy.ndarray -> Array of longitudes [rads] of nadir-scan soundings. 
            The dimension of the array is nscans x nfootprints, where nscans is the 
            number of scans in the returned data set and nfootprints is the number of 
            footprints in each scan.

    'latitudes': numpy.ndarray -> Array of latitudes [rads] of nadir-scan soundings. 
            The dimension of the array is nscans x nfootprints, where nscans is the 
            number of scans in the returned data set and nfootprints is the number of 
            footprints in each scan.

    'mid_times': list -> numpy.ndarray of numpy.datetime64 objects corresponding to 
            the middle of the scans in the returned data set. It represents UTC, and 
            its length is nscans. 

    'files': list -> List of file names containing the data. These are the paths to the 
            files containing the nadir-scan satellite data. Its length is nfiles. 

    'file_index': list -> List of indices pointing to the file (in 'files') containing 
            the data for a particular sounding. For example files[file_index[iscan]] is 
            the path to the file containing data for scan iscan in the returned data. 
            The length of the list is nscans. 

    'scan_index': list -> List of scan indices pointing to the scan index in the data file 
            indicating where to find the data. 

            For example, data[scan_index[iscan],:] in 
            file files[file_index[iscan]] is the data for scan iscan in an instance of 
            this class."""

    def __init__( self, get_data, longitudes, latitudes, mid_times, 
                 files, file_indices, scan_indices ): 
        """Longitudes and latitudes must be list-like objects in units of radians. 
        mid_times must be a list of timestandards.Time instances. files is a list of 
        strings; file_indices a list-like object of integers, as is scan_indices."""

        self.get_data = get_data
        self.longitudes = np.array( longitudes )
        self.latitudes = np.array( latitudes )
        self.mid_times = mid_times
        self.files = list( files )
        self.file_indices = np.array( file_indices )
        self.scan_indices = np.array( scan_indices )

    def __call__( self, iscan, ifootprint, **kwargs ): 
        """Retrieve a dictionary containing the radiance data for iscan and ifootprint 
        in this instance of ScanMetadata. The retrieved radiance data should correspond to 
        file self.files[iscan] with values data[self.scan_indices[iscan],self.footprint_indices[ifootprint]]. 
        The arguments iscan and ifootprint must be integers."""

        #  Check input. 

        integer_types = [ int, np.int8, np.int16, np.int32, np.int64 ]

        if not any( [ isinstance(iscan,t) for t in integer_types] ) or \
                not any( [ isinstance(ifootprint,t) for t in integer_types ] ) : 

            raise nadirSatelliteError( "InvalidArguments", 
                    "Arguments iscan and ifootprint must be integers." )

        file = self.files[ self.file_indices[ iscan ] ]
        scan_index = self.scan_indices[iscan]
        footprint_index = ifootprint

        ret = self.get_data( file, scan_index, footprint_index, **kwargs )

        return ret




"""metop_amsua.py

Authors: Meredith, Stephen Leroy
Contact: Stephen Leroy (sleroy@aer.com)
Date: March 5, 2024

This contains the definition of an ATMS instrument class on board 
a JPSS satellite --- including Suomi-NPP, JPSS-1 (NOAA-20), JPSS-2 
(NOAA-21), etc. --- with data hosted in NASA Earthdata DAACs. The class 
provided is named by class_name, which inherits the NadirSatelliteInstrument
class. 
"""

from netCDF4 import Dataset
import numpy as np
from datetime import datetime
import xarray 

from ..core.nadir_satellite import NadirSatelliteInstrument, ScanMetadata
from ..core.timestandards import Time, Calendar
from ..core.eumetsat import eumetsat_time_convention
from ..core.constants_and_utils import masked_dataarray, planck_blackbody, speed_of_light 

#  REQUIRED attributes 

instrument = "ATMS"
instrument_aliases = [ "ATMS", "atms", "JPSS ATMS" ]
valid_satellites = [ "Suomi-NPP", "JPSS-1" ]

#  REQUIRED methods of the class: 
#   - get_geolocations
#   - get_data


#  Parameters. 

fill_value = -1.0e20

#  Exception handling. 

class Error( Exception ): 
    pass

class nasa_jpss_error( Error ): 
    def __init__( self, message, comment ): 
        self.message = message
        self.comment = comment


#  Buffer for data files. 

open_data_file = { 'path': None, 'pointer': None }


class ATMS(NadirSatelliteInstrument):
    """JPSS/Suomi-NPP ATMS satellite instrument object, inheriting from 
    NadirSatelliteInstrument. This represents an ATMS scanner aboard a 
    Suomi-NPP/JPSS satellite.

    Parameters
    ------------
        name: str
            Name of the nadir satellite, drawn from awsgnssroutils.collocation.core.spacetrack.Satellites[:]['name']
        nasa_earthdata_access: collocation.core.nasa_earthdata.NASAEarthdata
            An object that interfaces with the NASA Earthdata DAACs. 
        spacetrack: collocation.core.spacetrack.Spacetrack
            Portal to the Spacetrack TLE data

    Attributes
    ------------
        name: str
            Name of the nadir satellite, drawn from awsgnssroutils.collocation.core.spacetrack.Satellites[:]['name']
        spacetrack_satellite: instance of spacetrack.SpacetrackSatellite
            Define the satellite, for access to TLEs
        xi: float
            MWR instrument maximum scan angle [radians]
        data: string
            Filepath to sounding data
        time_between_scans: float
            Time between cross-track scans [seconds]
        scan_points_per_line: int
            Number of individual footprints per cross-track scan
        scan_angle_spacing: float
            Angle between scan footprints [radians]
        eumetsat_access: collocation.core.eumetsat.EUMETSATDataStore
            An object that interfaces with the EUMETSAT Data Store. 
        spacetrack_satellite: collocation.core.spacetrack.SpacetrackSatellite
            TLE data for the satellite
    """

    def __init__(self, name, nasa_earthdata_access, spacetrack=None ):
        """Constructor for MetopAMSUA class."""

        self.instrument_name = instrument

        if name not in valid_satellites: 
            print( f'No {instrument} on satellite {name}. Valid satellites are ' + ", ".join( valid_satellites ) )
            self.status = "fail"

        else: 
            max_scan_angle = 52.63              # degrees
            time_between_scans = 8.0/3          # seconds
            scan_points_per_line = 96           # footprints in a scan
            scan_angle_spacing = 1.108          # degrees

            super().__init__( name, max_scan_angle, time_between_scans,
                    scan_points_per_line, scan_angle_spacing, spacetrack=spacetrack )

            self.nasa_earthdata_access = nasa_earthdata_access
            self.status = "success"

        pass

    def populate( self, timerange ): 
        """Populate (download) all JPSS ATMS data for this satellite in a 
        time range defined by timerange. timerange can be a 2-tuple/list of 
        instances of timestandards.Time or datetime.datetime with the 
        understanding that the latter is defined as UTC times."""

        self.nasa_earthdata_access.populate( self.satellite_name, 'atms', timerange )
        return

    def get_geolocations_from_file( self, filename ):
        """Get geolocation information from a single input file filename."""

        #  Define time standard for timing information in EUMETSAT Data Store 
        #  AMSU-A level 1a files. Possibilities are "utc", "tai", "gps". 

        #  Open file. 

        d = Dataset( filename, 'r' )

        #  Get dimensions. 

        nx = d.dimensions['xtrack'].size
        ny = d.dimensions['atrack'].size

        #  Get longitudes and latitudes. Convert to radians. 

        latitudes = np.deg2rad( d.variables['lat'][:] )
        longitudes = np.deg2rad( d.variables['lon'][:] )

        #  Get start and stop time of scans in file. 

        xtuples = d.variables['obs_time_utc'][:,int(nx/2),:]

        mid_times = []
        for iscan in range(ny): 
            if np.any( xtuples[iscan,0:8].mask ): 
                tt = None
            else: 
                tt = Time( utc=Calendar( *( xtuples[iscan,0:6] ) ) ) \
                        + xtuples[iscan,6]*1.0e-3 + xtuples[iscan,7]*1.0e-6 
            mid_times.append( tt )

        # mid_times = [ Time( utc=Calendar( *( xtuples[iscan,0:6] ) ) ) \
                # + xtuples[iscan,6]*1.0e-3 + xtuples[iscan,7]*1.0e-6 \
                # for iscan in range(ny) ]

        d.close()

        return { 'longitudes': longitudes, 'latitudes': latitudes, 'mid_times': mid_times }

    def get_geolocations( self, timerange ):
        """Load AMSU-A data from a Metop satellite as obtained from the EUMETSAT Data Store
        using collocation.core.eumetsat.EUMETSATDataStore.satellite.populate_metop_amsua. The 
        timerange is a two-element tuple/list containing instances of timestandards.Time 
        that prescribe the time range over which to obtain AMSU-A geolocations. An instance 
        of class ScanMetadata containing the footprint geolocations is returned upon successful 
        completion."""

        #  The EUMETSAT data is stored by orbit. In order to find all relevant AMSU-A 
        #  soundings, be sure to subtract one orbital period from the first time and 
        #  add one orbital period to the last time. 

        data_files = self.nasa_earthdata_access.get_paths( self.satellite_name, 'atms', timerange )
        gps0 = Time(gps=0)
        dt = ( timerange[0]-gps0, timerange[1]-gps0 )

        #  Loop over data files. Keep geolocations only for soundings within the timerange. 
        #  Initialize geolocation variables. 

        longitudes, latitudes, mid_times, scan_indices, file_indices = [], [], [], [], [] 

        for ifile, data_file in enumerate(data_files): 

            #  Read all geolocation information from data_file. 

            ret = self.get_geolocations_from_file( data_file )

            #  Convert mid_times to an np.ndarray of GPS times.  Find times for soundings 
            #  within the prescribed timerange. 

            xtimes = np.zeros( len(ret['mid_times']), dtype=np.float64 )
            mask = np.zeros( len(ret['mid_times']), dtype=np.byte )

            for i, t in enumerate( ret['mid_times'] ): 
                if t is None: 
                    xtimes[i] = -1001
                    mask[i] = 1
                else: 
                    xtimes[i] = t - gps0
                    mask[i] = 0

            file_gps_times = np.ma.masked_where( mask, xtimes )

            good = np.argwhere( np.logical_and( dt[0] <= file_gps_times, file_gps_times < dt[1] ) ).flatten()

            #  Keep only those soundings within the prescribed timerange. 

            if good.size > 0: 
                longitudes += [ ret['longitudes'][iy,:] for iy in good ]
                latitudes += [ ret['latitudes'][iy,:] for iy in good ]
                mid_times += [ Time(gps=file_gps_times[iy]) for iy in good ]
                scan_indices += list( good )
                file_indices += [ifile] * good.size

        #  Convert to ndarrays. 

        longitudes = np.array( longitudes )
        latitudes = np.array( latitudes )

        #  Generate output object. 

        ret = ScanMetadata( self.get_data, longitudes, latitudes, mid_times, data_files, file_indices, scan_indices )

        return ret

    def get_data( self, file, scan_index, footprint_index, 
                 longitude=None, latitude=None, time=None ): 
        """A function which returns nadir-scan satellite data for requested scan and 
        footprint indices within the file *file*. The function itself must be a 
        Method to fetch data for a nadir-scan satellite instrument corresponding to 
        a scalar integer indicating the scan number [0:nscans] and a scalar integer 
        indicating the footprint number [0:nfootprints].

        It returns an xarray.Dataset according taken from file *file* and the 
        data location within the file should correspond to scans 
        data[scan_index,footprint_index].

        Longitude (degrees), latitude (degrees), and time (timestandards.Time) will 
        be included in the returned xarray.Dataset if provided."""

        #  Check input. 

        integer_types = [ int, np.int8, np.int16, np.int32, np.int64 ]

        if not isinstance(file,str): 
            raise nasa_jpss_error( "InvalidArgument", "file argument must be type str" )

        if not any( [ isinstance(scan_index,t) for t in integer_types ] ): 
            raise nasa_jpss_error( "InvalidArgument", "scan_index argument must be an integer type" )

        if not any( [ isinstance(footprint_index,t) for t in integer_types ] ): 
            raise nasa_jpss_error( "InvalidArgument", "footprint_index argument must be an integer type" )

        #  Open data file. 

        global open_data_file

        if open_data_file['pointer'] is not None: 
            if open_data_file['path'] == file: 
                #  Data file already open. 
                d = open_data_file['pointer']
            else: 
                #  Close previously opened file. 
                open_data_file['pointer'].close()
                open_data_file['path'] = None
                open_data_file['pointer'] = None

        if open_data_file['path'] is None: 
            #  Open new file. 
            open_data_file['path'] = file
            open_data_file['pointer'] = Dataset( file, 'r' )
            d = open_data_file['pointer']

        dim_nscans, dim_nfootprints = d.dimensions['atrack'].size, d.dimensions['xtrack'].size
        nchannels = d.dimensions['channel'].size  

        #  Get data values. 

        brightness_temperature = d.variables['antenna_temp'][scan_index,footprint_index,:].flatten()

        #  Convert brightness tempereature to radiance. Convert radiances from 
        #  W m**-2 Hz**-1 ster**-1 to mW m**-2 (cm**-1)**-1 ster**-1. 

        frequencies = d.variables['center_freq'][:] * 1.0e6     #  Convert to Hz. 
        radiances = planck_blackbody( frequencies, brightness_temperature ) \
                * 1.0e3 * speed_of_light * 100.0 

        zenith = d.variables['sat_zen'][scan_index,footprint_index]

        #  Convert to np.ndarrays. 

        radiance_dataarray = masked_dataarray( radiances, fill_value, 
                dims=("channel",), 
                coords = { 'channel': np.arange(nchannels,dtype=np.int32)+1 } )
        radiance_dataarray.attrs.update( {
            'description': "Microwave radiance from ATMS instrument", 
            'units': "mW m**-2 (cm**-1)**-1 steradian**-1" } )

        zenith_dataarray = masked_dataarray( zenith, fill_value )
        zenith_dataarray.attrs.update( {
            'description': "Zenith angle from surface to satellite", 
            'units': "degrees" } )

        ds_dict = { 
            'data': radiance_dataarray, 
            'zenith': zenith_dataarray } 

        ds_attrs_dict = { 
            'satellite': self.satellite_name, 
            'instrument': self.instrument_name, 
            'data_file_path': file, 
            'scan_index': np.int16( scan_index ), 
            'footprint_index': np.int16( footprint_index ) } 

        if longitude is not None: 
            longitude_dataarray = xarray.DataArray( longitude )
            longitude_dataarray.attrs.update( { 
                'description': "Longitude of sounding, eastward", 
                'units': "degrees" } )
            ds_dict.update( { 'longitude': longitude_dataarray } )

        if latitude is not None: 
            latitude_dataarray = xarray.DataArray( latitude )
            latitude_dataarray.attrs.update( { 
                'description': "Latitude of sounding, northward", 
                'units': "degrees" } )
            ds_dict.update( { 'latitude': latitude_dataarray } )

        if time is not None: 
            ds_dict.update( { 'time': time.calendar("utc").isoformat(timespec="milliseconds")+"Z" } )

        ds = xarray.Dataset( ds_dict )
        ds.attrs.update( ds_attrs_dict )

        return ds




instruments = {}

from .atms import ATMS, instrument, valid_satellites
instruments.update( { instrument: { 'class': ATMS, 'valid_satellites': valid_satellites } } )

from .amsua import AMSUA, instrument, valid_satellites
instruments.update( { instrument: { 'class': AMSUA, 'valid_satellites': valid_satellites } } )

satellites = {}
for instrument, value in instruments.items(): 
    for satellite in value['valid_satellites']: 
        if satellite not in satellites.keys(): 
            satellites.update( { satellite: [] } )
        satellites[satellite].append( instrument )




"""Utilities for reading Space-Track TLE data."""

#  Imports. 

import os, re, json, stat
import numpy as np
import requests
from datetime import datetime, timedelta
from .timestandards import Time, Calendar
from .constants_and_utils import defaults_file

#  Definitions. 

root_path_variable = "spacetrack_data_root"
username_variable = "spacetrack_username"
password_variable = "spacetrack_password"

#  REST API querying command parameters. 

uri_base = "https://www.space-track.org"
authenticate_command = "ajaxauth/login"

#  There are three arguments to str.format in download_command: 
#   (1) 5-digit NORAD catalogue number specifying the satellite
#   (2) Begin date of the epoch range in the format YYYY-MM-DD
#   (3) End date of the epoch range in the format YYYY-MM-DD

download_command = "basicspacedata/query/class/tle/NORAD_CAT_ID/{:}/EPOCH/{:}--{:}/" + \
        "orderby/NORAD_CAT_ID%20asc/format/tle/emptyresult/show"

#  Parameters. 

spacetrack_time_convention = "utc"
gps0 = Time( gps=0 )

#  Table of satellites. 

Satellites = [ { 'name': "Metop-A", 'norad_number': 29499 }, 
              { 'name': "Metop-B", 'norad_number': 38771 }, 
              { 'name': "Metop-C", 'norad_number': 43689 }, 
              { 'name': "Suomi-NPP", 'norad_number': 37849 }, 
              { 'name': "JPSS-1", 'norad_number': 43013 }, 
              { 'name': "JPSS-2", 'norad_number': 54234 }, 
              { 'name': "Aqua", 'norad_number': 27424 } ]


#  Exception handling. 

class Error( Exception ): 
    pass

class spacetrackError( Error ): 
    def __init__( self, message, comment ): 
        self.message = message
        self.comment = comment


def setdefaults( root_path=None, spacetracklogin=None ):
    """This method sets the default root path for Spacetrack TLE data. If the 
    keyword spacetracklogin is given, it must be a 2-element tuple/list 
    containing the user's username and password for the user's Space-Track 
    account."""

    ret = { 'status': None, 'messages': [], 'comments': [] }

    new_defaults = {}

    #  Update data root path and create the directory. 

    if isinstance(root_path,str): 
        path = os.path.abspath( os.path.expanduser( root_path ) ) 
        new_defaults.update( { root_path_variable: path } )
        os.makedirs( path, exist_ok=True )

    #  Update username and password. 

    if spacetracklogin is not None: 

        if isinstance(spacetracklogin,tuple) or isinstance(spacetracklogin,list): 
            if len(spacetracklogin) == 2: 
                new_defaults.update( { username_variable: spacetracklogin[0], 
                          password_variable: spacetracklogin[1] } ) 

            else:
                ret['status'] = "fail"
                ret['messages'].append( "InvalidArgument" )
                ret['comments'].append( "spacetracklogin must have length 2" )
                return ret

        else:
            ret['status'] = "fail"
            ret['messages'].append( "InvalidArgument" )
            ret['comments'].append( "spacetracklogin must be a tuple or a list" )
            return ret

    #  Get old defaults. 

    if os.path.exists( defaults_file ):
        with open( defaults_file, 'r' ) as f:
            defaults = json.load( f )
    else: 
        defaults = {}

    #  Update with new defaults. 

    defaults.update( new_defaults )

    #  Write to defaults file. 

    with open( defaults_file, 'w' ) as f:  
        json.dump( defaults, f, indent="  " ) 
    os.chmod( defaults_file, stat.S_IRUSR | stat.S_IWUSR )

    #  Done. 

    ret['status'] = "success"

    return ret


def checkdefaults(): 
    """Check that all of the defaults needed for execution have been set."""

    ret = { 'status': None, 'messages': [], 'comments': [] }

    #  Read defaults file. 

    with open( defaults_file, 'r' ) as f: 
        defaults = json.load( f )

    #  Check that the data root path has been set. 

    keys = defaults.keys()

    if root_path_variable not in keys: 
        ret['status'] = "fail"
        ret['messages'].append( "MissingEarthdataRoot" )
        ret['comments'].append( 'Missing data root for Space-Track; be certain to run "rotcol setdefaults spacetrack --dataroot ..."' )
        return ret

    #  Check that the username and password have been set. 

    if username_variable not in keys and password_variable not in keys: 
        ret['status'] = "fail"
        ret['messages'].append( "MissingSpacetrackCredentials" )
        ret['comments'].append( 'Missing credentials for Space-Track; be certain to run ' + \
                '"rotcol setdefaults spacetrack --username ... --password ..."' )
        return ret

    #  Done. 

    ret['status'] = "success"
    return ret


class SpacetrackSatellite(): 
    """This class contains the TLE data, as obtained from Spacetrack, for one 
    satellite."""

    def __init__( self, Spacetrack_instance, norad_number ): 
        """Create an instance of class SpacetrackSatellite. It will contain the 
        TLE data for a satellite identified by its NORAD catalogue number 
        (norad_number). The first argument is the instance of class Spacetrack 
        from which this class is drawn."""

        if not isinstance( Spacetrack_instance, Spacetrack ): 
            raise spacetrackError( "InvalidArgument", "First argument to " + \
                    "SpacetrackSatellite not an instance of class Spacetrack" )

        #  Initialize. 

        self.status = None
        self.nrecs = None
        self.datetimes = None
        self.tles = None

        self.spacetrack = Spacetrack_instance
        self.norad_number = norad_number
        norad_number_string = f'{norad_number:05d}'

        if norad_number_string not in self.spacetrack.catalogue.keys(): 
            self.status = "success"
            self.nrecs = 0
            self.datetimes = []
            self.tles = []
        else: 
            self.build_instance()

        pass

    def build_instance( self ): 
        """Reconstruct the data in this instance."""

        data = {}

        norad_number_string = f'{self.norad_number:05d}'
        for file in self.spacetrack.catalogue[norad_number_string]: 

            f = open( file, 'r' )
            lines = f.readlines()
            f.close()

            if len(lines) % 2 == 1: 
                raise spacetrackError( "InvalidData", f'file {file} has an odd number of lines' )

            for iline in range(0,len(lines),2): 

                tle_line1 = lines[iline]
                tle_line2 = lines[iline+1]

                if tle_line1[0] != "1": 
                    raise spacetrackError( "InvalidData", 
                            f'Line {iline+1} in file {file} should be the first line of a TLE but is not' )

                if tle_line2[0] != "2": 
                    raise spacetrackError( "InvalidData", 
                            f'Line {iline+2} in file {file} should be the second line of a TLE but is not' )

                #  Get the time of the record. 

                year = 2000 + int( tle_line1[18:20] )
                doy = int( tle_line1[20:23] )
                fday = float( tle_line1[23:32] )

                time = Time( **{ spacetrack_time_convention: Calendar(year,1,1) } ) + 86400 * ( doy + fday - 1 )
                rec = { 'time': time, 'tle': ( tle_line1, tle_line2 ) }
                data.update( { time.calendar(spacetrack_time_convention).isoformat(timespec="seconds"): rec } )

        #  Sort the records. 

        sorted_key_list = sorted( list( data.keys() ) )

        #  Create data contents. 

        self.nrecs = len( sorted_key_list )
        self.datetimes = [ data[key]['time'] for key in sorted_key_list ]
        self.tles = [ data[key]['tle'] for key in sorted_key_list ]

        self.status = "success"

        return

    def select( self, timerange, extend=None ): 
        """Generate a list of TLEs. Each element of output will be a 2-tuple containing 
        the two lines of a TLE. All TLEs in the list must fall within a range of 
        datetimes defined by the two-element tuple/list datetimerange. 

        Arguments: 

        timerange       A two-element tuple/list containing two instances of timestandards.Time. 
                        The first element defines the start time of the TLE list to be 
                        generated; the second defines the end time of the TLE list. 

        extend          A two-element tuple/list of integers defining the how far to extend 
                        the list of TLEs before the datetimerange and after the datetimerange. 

        Example: 

        data = SpacetrackSatellite.select( 
                timerange = [ timestandards.Time( utc=timestandards.Calendar(2023,6,7) ), 
                                timestandards.Time( utc=timestandards.Calendar(2023,6,8) ) ], 
                extend = [0,1] )

        The statement generates a list of TLEs for the satellite defined by SpacetrackSatllite
        for the day of June 7, 2023, and appends to the list the first TLE for the succeeding 
        day."""

        if not isinstance(timerange,list) and not isinstance(timerange,tuple): 
            raise spacetrackError( "InvalidArgument", 'timerange must be a tuple or list' )

        if len( timerange ) != 2: 
            raise spacetrackError( "InvalidArgument", 'timerange must containing two elements' )

        if not isinstance(timerange[0],Time) or not isinstance(timerange[1],Time): 
            raise spacetrackError( "InvalidArgument", 'timerange must consist of timestandards.Time instances' )

        if extend is not None: 

            if not isinstance(extend,list) and not isinstance(extend,tuple): 
                raise spacetrackError( "InvalidArgument", 'extend must be a tuple or list' )

            if len( extend ) != 2: 
                raise spacetrackError( "InvalidArgument", 'extend must containing two elements' )

            if not isinstance(extend[0],int) or not isinstance(extend[1],int): 
                raise spacetrackError( "InvalidArgument", 'extend must consist of two int' )

        #  Extract elements. 

        dtr = np.array( [ dt - gps0 for dt in timerange ] )
        datetimes_gps = np.array( [ dt - gps0 for dt in self.datetimes ] )
        ii = np.argwhere( np.logical_and( dtr[0] <= datetimes_gps, datetimes_gps < dtr[1] ) ).squeeze()

        #  Extend? 

        if extend is None: 
            imin = ii[0]
            imax = ii[-1]
        else: 
            imin = max( [ 0, ii[0]-extend[0] ] )
            imax = min( [ self.nrecs-1, ii[-1]+extend[1] ] )

        #  Extract TLEs. 

        ret = [ self.tles[i] for i in range(imin,imax+1) ]

        #  Done. 

        return ret


    def nearest( self, time ): 
        """Get the TLE that is nearest in time (instance of timestandards.Time). The two lines of the 
        TLE are returned as a 2-tuple if successful. If unsuccessful, None is returned."""

        ret = { 'status': None, 'messages': [], 'comments': [] }

        if not isinstance( time, Time ): 
            raise spacetrackError( "InvalidArgument", "The argument must be an instance of timestandards.Time" )

        #  Iterate to assure update of local TLE data (if necessary). 

        for iteration in range(2): 

            datetimes_gps = np.array( [ dt-gps0 for dt in self.datetimes ] )
            time_gps = time - gps0

            if datetimes_gps.size == 0: 
                ret_tles = None
                ret_spacetrack = self.spacetrack.download_data( self.norad_number, time.calendar("utc").datetime() )
                self.build_instance()
                continue

            i = np.argmin( np.abs( datetimes_gps - time_gps ) ) 

            #  Check for valid return value. The nearest time must be better than 12 hours. 

            if np.abs( datetimes_gps[i] - time_gps ) < 12 * 3600: 
                ret_tles = self.tles[i]
                break 
            elif iteration==0: 
                ret_tles = None
                ret_spacetrack = self.spacetrack.download_data( self.norad_number, time.calendar("utc").datetime() )
                ret['messages'] += ret_spacetrack['messages']
                ret['comments'] += ret_spacetrack['comments']
                self.build_instance()
            else: 
                ret_tles = None


        if ret_tles is None: 
            ret['status'] = "fail"
            ret['messages'].append( "NoOrbitData" )
            ret['comments'].append( "No orbit data available at Space-Track.org for " + \
                    time.calendar("utc").isoformat( timespec="seconds" ) )
        else: 
            ret['status'] = "success"
            ret.update( { 'data': ret_tles } )

        return ret



class Spacetrack(): 
    """This is a portal to handling TLE data obtained from Spacetrack."""

    def __init__( self ): 
        """Create a portal to access TLE downloaded from spacetrack.org. Data must be 
        strict two-line element files, with each file containing data for only one 
        satellite."""

        with open( defaults_file, 'r' ) as f: 
            defaults = json.load( f )
            self.data_root = defaults[root_path_variable]

        if not os.path.isdir( self.data_root ): 
            raise spacetrackError( "PathNotFound", f'"{root_path_variable} does not exist or is not a directory' )

        #  Establish an authenticated session. 

        myauth = { 'identity': defaults[username_variable], 'password': defaults[password_variable] }
        self.session = requests.Session()
        self.session.request( "post", "https://www.space-track.org/ajaxauth/login", data=myauth )

        #  Initiate the catalogue of TLE data. 

        self.catalogue = {}
        self.build_catalogue()


    def build_catalogue( self ): 
        """Build a catalogue of the TLE data currently available on the local 
        file system."""

        #  Scan directory hierarchy looking for data files. 

        for root, dirs, files in os.walk( self.data_root ): 

            files.sort()
            dirs.sort()

            for file in files: 
                m = re.search( "^sat(\d{5}).*\.txt$", file )
                if not m: continue

                #  Read first line of the file to get the satellite NORAD number. 

                file_path = os.path.join( root, file )
                f = open( file_path, 'r' )
                line = f.readline()
                norad_number_string = line[2:7]
                f.close()

                #  Catalog the file. 

                if norad_number_string not in self.catalogue.keys(): 
                    self.catalogue.update( { norad_number_string: [] } )
                self.catalogue[norad_number_string].append( file_path )

            pass 

        #  Done. 

        pass

    def satellite( self, generic ): 
        """Create an instance of class SpacetrackSatellite given either a 
        NORAD catalogue number (int) or the name of the satellite (str) as 
        given in the table Satellites."""

        #  Define the NORAD catalogue number of the satellite. 
        
        norad_number = 0
        
        if isinstance( generic, int ): 
            norad_number = generic
        elif isinstance( generic, str ): 
            s = [ sat for sat in Satellites if sat['name'] == generic ]
            if len(s) == 1: 
                norad_number = s[0]['norad_number']
                
        #  Convert to string. 
        
        if norad_number != 0: 
            norad_number_string = f'{norad_number:05d}'
        else: 
            norad_number_string = ""
        
        #  Generate the Spacetrack satellite, load TLE data if available. 
        
        ret = SpacetrackSatellite( self, norad_number )

        return ret

    def download_data( self, norad_number, time ): 
        """Download TLEs for a satellite defined by its NORAD number 
        (norad_number) and a time. The default is to obtain one month's 
        worth of data. 

        norad_number - Either a string or an integer representing 
                the NORAD number of the satellite whose TLEs should 
                be obtained from SpaceTrack. 

        time - an instance of datetime.datetime for the time for which 
                TLE is desired. A full month's worth of data will be 
                obtained."""

        #  Initialize return structure. 

        ret = { 'status': None, 'messages': [], 'comments': [] }

        #  Check input. 

        if not isinstance(time,datetime): 
            ret['status'] = "fail"
            ret['messages'].append( "InvalidArgument" )
            ret['comments'].append( "The time must be an instance of datetime.datetime" )
            return ret

        if isinstance(norad_number,str): 
            satID = norad_number[-5:]
        elif isinstance(norad_number,int): 
            satID = f'{norad_number:05d}'
        else: 
            ret['status'] = "fail"
            ret['messages'].append( "InvalidArgument" )
            ret['comments'].append( "The time must be an instance of datetime.datetime" )
            return ret

        #  Define the first and last time of a month. 

        t0 = datetime( year=time.year, month=time.month, day=1 )
        tt = t0 + timedelta(days=31)
        t1 = datetime( year=tt.year, month=tt.month, day=1 )

        #  Post the request. 

        tf = "%Y-%m-%d"
        request_uri = "/".join( [ uri_base, download_command.format( satID, t0.strftime(tf), t1.strftime(tf) ) ] )
        resp = self.session.get( request_uri )

        if resp.status_code != 200: 
            ret['status'] = "fail"
            ret['messages'].append( "FailedDownload" )
            ret['comments'].append( f'failed request_uri = "{request_uri}"' )
            return ret

        #  Write to output file. 

        path = os.path.join( self.data_root, f'sat{satID}', f'sat{satID}_'+time.strftime("%Y-%m")+".txt" )
        os.makedirs( os.path.dirname(path), exist_ok=True )
        with open(path,'w') as f: 
            for line in resp.iter_lines(): 
                f.write( line.decode() + "\n" )

        #  Rebuild catalogue. 

        self.build_catalogue()

        #  Done. 

        ret['status'] = "success"

        return ret


if __name__ == "__main__": 

    data_dir = "../../Data/TLEs"
    cs = Spacetrack( data_dir )

    #  Select a satellite. 

    metopb = cs.satellite(43689)
    tle_data = metopb.select( ( datetime(2022,8,1), datetime(2022,9,1) ), extend=(1,1) )

    pass




"""nasa_earthdata.py


Author: Stephen Leroy (sleroy@aer.com)
Date: March 1, 2024

This module contains classes and methods useful for accessing the NASA 
DAAC.  Note that you'll have to create an account on the NASA Earthdata 
server."""


import os, re, stat, json, subprocess
import earthaccess
import requests, netrc, boto3
from platform import system
from datetime import datetime, timedelta, timezone
from .timestandards import Time
from .constants_and_utils import defaults_file

#  Definitions. 

#  ATMS pointers are for version 3 level 1B radiances. 
#  CrIS pointers are for "full spectral resolution, version 3" level 1B radiances. 

Satellites = { 
        'Aqua': { 
                 'aliases': [] }, 
        'Suomi-NPP': { 
                'aliases': [ "Suomi-NPP", "SNPP" ], 
                'atms': "10.5067/FCXKUUE9VCLN", 
                'cris': "10.5067/ZCRSHBM5HB23"
        }, 
        'JPSS-1': { 
                'aliases': [ "JPSS-1", "NOAA-20" ], 
                'atms': "10.5067/MUNII2DHSSY3", 
                'cris': "10.5067/LVEKYTNSRNKP" 
        }, 
        'JPSS-2': {
                'aliases': [ "JPSS-2", "NOAA-21" ] 
        }
    }

HOME = os.path.expanduser( "~" )
root_path_variable = "nasa_earthdata_root"
earthdata_machine = "urs.earthdata.nasa.gov"
time_limit = timedelta( seconds=3600 )

#  String parsing. 

file_search_string = "^SNDR\..*(\d{8}T\d{4})\.m.*\.nc$"
time_parse_string = "%Y%m%dT%H%M" 

#  Establist the netrc file name. 

if system() == "Windows": 
    netrc_file = os.path.join( HOME, "_netrc" )
else: 
    netrc_file = os.path.join( HOME, ".netrc" )

#  Define metadata on the Earthdata DAACs. 

earthdata_daacs = {
        'podaac': { 'endpoint': "https://archive.podaac.earthdata.nasa.gov/s3credentials", 
                    'region': "us-west-2" }, 
        'gesdisc': { 'endpoint': "https://data.gesdisc.earthdata.nasa.gov/s3credentials", 
                    'region': "us-west-2" } }

#  Exception handling. 

class Error( Exception ): 
    pass 

class earthdataError( Error ): 
    def __init__( self, message, comment ): 
        self.message = message
        self.comment = comment 


def setdefaults( root_path=None, earthdatalogin=None ): 
    """This method sets the default root path for NASA GDAAC downloads.
    Optionally, it also allows you to register your Earthdata username and 
    password in your home .netrc (if you haven't done so already).

    The Earthdata username and password can be given as a 2-tuple 
    (username,password) in optional argument earthdatalogin."""

    ret = { 'status': None, 'messages': [], 'comments': [] }

    #  Set root data path. Create the directory. 

    if isinstance(root_path,str): 

        if os.path.exists( defaults_file ): 
            with open( defaults_file, 'r' ) as f: 
                defaults = json.load( f )
        else: 
            defaults = {}

        path = os.path.abspath( os.path.expanduser( root_path ) ) 
        defaults.update( { root_path_variable: path } )
        os.makedirs( path, exist_ok=True )

        #  Write to defaults file. 

        with open( defaults_file, 'w' ) as f: 
            json.dump( defaults, f, indent="  " )
        os.chmod( defaults_file, stat.S_IRUSR | stat.S_IWUSR )

    #  Write Earthdata login username and password to .netrc file if 
    #  provided. 

    if earthdatalogin is not None: 
        if isinstance(earthdatalogin,tuple) or isinstance(earthdatalogin,list): 
            if len(earthdatalogin) == 2: 

                if os.path.exists( netrc_file ): 
                    with open( netrc_file, 'r' ) as f: 
                        lines = f.readlines()
                else: 
                    lines = []
                
                #  Catalog lines according to machine. 

                catalog = {}
                for line in lines: 
                    m = re.search( "^machine\s+(\S+)", line )
                    if m: 
                        catalog.update( { m.group(1): line.strip() } )

                #  Update catalog to include earthdata. 

                catalog.update( { earthdata_machine: \
                        "machine {:} login {:} password {:}".format( earthdata_machine, *earthdatalogin ) 
                        } )

                #  Write new netrc. 

                with open( netrc_file, 'w' ) as f: 
                    for machine, line in catalog.items(): 
                        f.write( line + "\n" )

                os.chmod( netrc_file, stat.S_IRUSR | stat.S_IWUSR )

            else: 

                ret['status'] = "fail"
                ret['messages'].append( "InvalidArgument" )
                ret['comments'].append( "earthdatalogin must have length 2" )
                return ret

        else: 

            ret['status'] = "fail"
            ret['messages'].append( "InvalidArgument" )
            ret['comments'].append( "earthdatalogin must be a tuple or a list" )
            return ret

    #  Done. 

    ret['status'] = "success" 

    return ret


def checkdefaults(): 
    """Check that all of the defaults needed for execution have been set."""

    ret = { 'status': None, 'messages': [], 'comments': [] }

    #  Check to see that the Earthdata machine has been defined in .netrc. 

    if not os.path.exists( netrc_file ): 
        ret['status'] = "fail"
        ret['messages'].append( "MissingNetrcFile" )
        ret['comments'].append( 'The .netrc file does not exists; be certain to run ' + \
                '"rotcol setdefaults earthdata ..."' )
        return ret

    #  Check to see that the earthdata_machine entry exists in the netrc file. 

    with open( netrc_file, 'r' ) as f: 
        lines = f.readlines()

    for line in lines: 
        m = re.search( "^machine\s+(\S+)", line )
        if m: 
            machine = m.group(1)
            if machine == earthdata_machine: 
                break
        else: 
            machine = None

    if machine != earthdata_machine: 
        ret['status'] = "fail"
        ret['messages'].append( "MissingEarthdataMachine" )
        ret['comments'].append( 'The .netrc file does not contain an entry for NASA Earthdata; be ' + \
                'certain to run "rotcol setdefaults earthdata --username ... --password ..."' )
        return ret

    #  Check for existence of package defaults file. 

    if not os.path.exists( defaults_file ): 
        ret['status'] = "fail"
        ret['messages'].append( "MissingDefaultsFile" )
        ret['comments'].append( 'Missing defaults file; be certain to run "rotcol setdefaults ..."' )
        return ret

    #  Read defaults file. 

    with open( defaults_file, 'r' ) as f: 
        defaults = json.load( f )

    #  Check that the data root path has been set. 

    if root_path_variable not in defaults.keys(): 
        ret['status'] = "fail"
        ret['messages'].append( "MissingEarthdataRoot" )
        ret['comments'].append( 'Missing data root for NASA Earthdata; be certain to run ' + \
                '"rotcol setdefaults earthdata --dataroot ..."' )
        return ret

    #  Done. 

    ret['status'] = "success"
    return ret


class NASAEarthdata(): 
    """Class to handle interaction with NASA DAACs."""

    def __init__( self ): 
        """Create an instance of EUMETSATDataStore. data_root should be an 
        absolute path to the storage location for downloaded EUMETSAT Data
        Store data."""

        #  Configure credentials. 

        self.earthaccess = earthaccess.login( strategy="netrc", persist=True )

        #  Get root data path. 

        with open( defaults_file, 'r' ) as f: 
            defaults = json.load( f )
            self.data_root = defaults[root_path_variable]
            os.makedirs( self.data_root, exist_ok=True )

        #  Initialize inventory. 

        self.inventory = {}
        self.regenerate_inventory()

        return

    def regenerate_inventory( self ): 
        """Create an inventory of the files available on the local file system 
        with ATMS data (as obtained from GES DISC."""

        #  Get list of satellites. 

        instruments = [ p for p in os.listdir( self.data_root ) if os.path.isdir( os.path.join( self.data_root, p ) ) ]

        #  Loop over instruments. 

        for instrument in instruments: 

            data_root = os.path.join( self.data_root, instrument )

            #  Initialize inventory. 

            if os.path.isdir( data_root ): 
                satellites = [ p for p in os.listdir( data_root ) if p in Satellites.keys() \
                        and os.path.isdir( os.path.join( data_root, p ) ) ]
            else: 
                satellites = []

            if instrument not in self.inventory: 
                self.inventory.update( { instrument: {} } )

            #  Loop over satellites. 

            for sat in satellites: 

                for root, subdirs, files in os.walk( os.path.join( data_root, sat ) ): 
                    subdirs.sort()
                    files.sort()

                    if sat not in self.inventory[instrument]: 
                        self.inventory[instrument].update( { sat: [] } )

                    for file in files: 
                        m = re.search( file_search_string, file )
                        if m is None: 
                            continue
                        t1 = Time( utc = datetime.strptime( m.group(1), time_parse_string ) ) 
                        t2 = t1 + 6 * 60

                        rec = { 'satellite': sat, 'path': os.path.join( root, file ), 'timerange': ( t1, t2 ) }
                        self.inventory[instrument][sat].append( rec )

        return

    def get_paths( self, satellite, instrument, timerange ): 
        """Return a listing of the paths to JPSS ATMS data files given a 
        satellite name and a time range. The timerange is a two-element 
        tuple/list with instances of timestandards.Time or datetime.datetime. 
        If it is the latter, then the datetime elements are understood to be 
        UTC."""

        #  Check input. Interpret datetime.datetime as timestandards.Time instances 
        #  if necessary. 

        if satellite not in Satellites.keys(): 
            raise earthdataError( "InvalidArgument", "The satellite must be one of " + \
                    ", ".join( list( Satellites.keys() ) ) )

        elif instrument not in Satellites[satellite].keys() or instrument == "aliases": 
            raise earthdataError( "InvalidArgument", 
                    f"The instrument {instrument} is not defined for satellite {satellite}" )

        if not isinstance( timerange, tuple ) and not isinstance( timerange, list ): 
            raise earthdataError( "InvalidArgument", "timerange must be a tuple/list of two elements" )

        if len( timerange ) != 2: 
            raise earthdataError( "InvalidArgument", "timerange must be a tuple/list of two elements" )

        if isinstance( timerange[0], datetime ) and isinstance( timerange[1], datetime ): 
            _timerange = [ Time( utc=timerange[i] ) for i in range(2) ]
        elif isinstance( timerange[0], Time ) and isinstance( timerange[1], Time ): 
            _timerange = timerange
        else: 
            raise earthdataError( "InvalidArgument", "The elements of timerange must both be " + \
                    "datetime.datetime or timestandards.Time" )

        ret = []
        if instrument in self.inventory.keys(): 
            if satellite in self.inventory[instrument].keys(): 
                ret = sorted( [ rec['path'] for rec in self.inventory[instrument][satellite] if \
                        rec['timerange'][0] <= _timerange[1] and rec['timerange'][1] >= _timerange[0] ] )

        return ret

    def populate( self, satellite, instrument, timerange ): 
        """Download SNPP, JPSS ATMS data that fall within a timerange. 

        * satellite must be one of Satellites.keys(). 
        * instrument is one of 'atms', 'cris'. 
        * timerange is a 2-element tuple/list of instances of timestandards.Time 
          or instances of datetime.datetime defining the range of times over which 
          to retrieve data. If they are instances of datetime.datetime, then the 
          convention is that they are both UTC."""

        #  Check input. Interpret datetime.datetime as timestandards.Time instances 
        #  if necessary. 

        if satellite not in Satellites.keys(): 
            raise earthdataError( "InvalidArgument", "The satellite must be one of " + \
                    ", ".join( list( Satellites.keys() ) ) )

        elif instrument not in Satellites[satellite].keys() or instrument == "aliases": 
            raise earthdataError( "InvalidArgument", 
                    f"The instrument {instrument} is not defined for satellite {satellite}" )

        if not isinstance( timerange, tuple ) and not isinstance( timerange, list ): 
            raise earthdataError( "InvalidArgument", "timerange must be a tuple/list of two elements" )

        if len( timerange ) != 2: 
            raise earthdataError( "InvalidArgument", "timerange must be a tuple/list of two elements" )

        if isinstance( timerange[0], datetime ) and isinstance( timerange[1], datetime ): 
            _timerange = [ Time( utc=timerange[i] ) for i in range(2) ]
        elif isinstance( timerange[0], Time ) and isinstance( timerange[1], Time ): 
            _timerange = timerange
        else: 
            raise earthdataError( "InvalidArgument", "The elements of timerange must both be " + \
                    "datetime.datetime or timestandards.Time" )

        #  Query the local and remote inventories. 

        local_inventory = self.get_paths( satellite, instrument, _timerange )
        etimerange = [ _timerange[0], _timerange[1] + 86400 ]
        temporal = tuple( [ t.calendar("utc").datetime().strftime("%Y-%m-%d") for t in etimerange ] ) 
        
        remote_inventory = earthaccess.search_data( doi=Satellites[satellite][instrument], temporal=temporal )

        #  Get a listing of data files that are available at Earthdata but not in the local inventory. 

        local_basenames = [ os.path.basename( p ) for p in local_inventory ]

        get = []
        for p in remote_inventory: 
            basename = os.path.basename( p.data_links()[0] )
            if basename in local_basenames: 
                continue
            m = re.search( file_search_string, basename )
            t = Time( utc=datetime.strptime( m.group(1), time_parse_string ) )
            if t+360 >= _timerange[0] and t <= _timerange[1]: 
                get.append( p )

        #  Get data files that we don't yet have. 

        if len( get ) > 0: 

            earthaccess.download( get, "tmp" )
            files = sorted( [ os.path.join( "tmp", f ) for f in os.listdir( "tmp" ) ] )

            for file in files: 

                #  Parse file name for time of granule. 

                m = re.search( file_search_string, os.path.basename(file) )
                dt = datetime.strptime( m.group(1), time_parse_string )

                #  Define local path for file. 

                lpath = os.path.join( self.data_root, instrument, satellite, 
                            f'{dt.year:02d}', f'{dt.month:02d}', f'{dt.day:02d}', 
                            os.path.basename( file ) )

                #  Move file. 

                os.makedirs( os.path.dirname( lpath ), exist_ok=True )
                os.link( file, lpath )
                os.unlink( file )

        #  Regenerate inventory. 

        self.regenerate_inventory()

        return 

